"""Class that represents ``[vak.train]`` table of configuration file."""

from attrs import converters, define, field, validators
from attrs.validators import instance_of

from ..common.converters import bool_from_str, expanded_user_path
from .dataset import DatasetConfig
from .model import ModelConfig
from .trainer import TrainerConfig

REQUIRED_KEYS = (
    "dataset",
    "model",
    "root_results_dir",
    "trainer",
)


@define
class TrainConfig:
    """Class that represents ``[vak.train]`` table of configuration file.

    Attributes
    ----------
    model : vak.config.ModelConfig
        The model to use: its name,
        and the parameters to configure it.
        Must be an instance of :class:`vak.config.ModelConfig`
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    batch_size : int
        number of samples per batch presented to models during training.
    root_results_dir : str
        directory in which results will be created.
        The vak.cli.train function will create
        a subdirectory in this directory each time it runs.
    dataset : vak.config.DatasetConfig
        The dataset to use: the path to it,
        and optionally a path to a file representing splits,
        and the name, if it is a built-in dataset.
        Must be an instance of :class:`vak.config.DatasetConfig`.
    trainer : vak.config.TrainerConfig
        Configuration for :class:`lightning.pytorch.Trainer`.
        Must be an instance of :class:`vak.config.TrainerConfig`.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model.
        Default is None, in which case a new model is initialized.
    frames_standardizer_path : str
        path to a saved :class:`vak.transforms.FramesStandardizer` object used to standardize (normalize) frames.
        If spectrograms were normalized and this is not provided, will give
        incorrect results. Default is None.
    """

    # required
    model = field(
        validator=instance_of(ModelConfig),
    )
    num_epochs = field(converter=int, validator=instance_of(int))
    batch_size = field(converter=int, validator=instance_of(int))
    root_results_dir = field(converter=expanded_user_path)
    dataset: DatasetConfig = field(
        validator=instance_of(DatasetConfig),
    )
    trainer: TrainerConfig = field(
        validator=instance_of(TrainerConfig),
    )

    results_dirname = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )

    standardize_frames = field(
        converter=bool_from_str,
        validator=validators.optional(instance_of(bool)),
        default=False,
    )

    num_workers = field(validator=instance_of(int), default=2)
    shuffle = field(
        converter=bool_from_str, validator=instance_of(bool), default=True
    )

    val_step = field(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    ckpt_step = field(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    patience = field(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    checkpoint_path = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )
    frames_standardizer_path = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )

    @classmethod
    def from_config_dict(cls, config_dict: dict) -> "TrainConfig":
        """Return :class:`TrainConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using key ``train``,
        i.e., ``TrainConfig.from_config_dict(config_dict['train'])``."""
        for required_key in REQUIRED_KEYS:
            if required_key not in config_dict:
                raise KeyError(
                    "The `[vak.train]` table in a configuration file requires "
                    f"the option '{required_key}', but it was not found "
                    "when loading the configuration file into a Python dictionary. "
                    "Please check that the configuration file is formatted correctly."
                )
        config_dict["model"] = ModelConfig.from_config_dict(
            config_dict["model"]
        )
        config_dict["dataset"] = DatasetConfig.from_config_dict(
            config_dict["dataset"]
        )
        config_dict["trainer"] = TrainerConfig(**config_dict["trainer"])
        return cls(**config_dict)



"""A dataset class used for neural network models with the
frame classification task, where the source data consists of audio signals
or spectrograms of varying lengths.

Unlike :class:`vak.datasets.frame_classification.InferDatapipe`,
this class does not return entire samples
from the source dataset.
Instead each paired samples :math:`(x_i, y_i)`
returned by this dataset class consists of
a window :math:`x_i` of fixed length
:math:`w` from the underlying data ``X`` of total length :math:`T`.
Each :math:`y_i` is a vector of the same size :math:`w`, containing
an integer class label for each *frame* in the window :math:`x_i`.
The entire dataset consists of some number of windows
:math:`I` determined by a ``stride`` parameter :math:`s`,
:math:`I = (T - w) / s`.
"""

from __future__ import annotations

import pathlib

import numpy as np
import numpy.typing as npt
import pandas as pd

from ...transforms import FramesStandardizer
from ...transforms.defaults.frame_classification import TrainItemTransform
from . import constants, helper
from .metadata import Metadata


def get_window_inds(n_frames: int, window_size: int, stride: int = 1):
    """Get indices of windows for a :class:`TrainDatapipe`,
    given the number of frames in the dataset,
    the window size, and the stride.

    This function is used by :class:`TrainDatapipe`
    to compute the indices of windows in the dataset.
    The length of the vector of indices it returns
    is the number of windows in the dataset,
    i.e., the number of samples.

    Parameters
    ----------
    n_frames : int
    window_size : int
    stride : int

    Returns
    -------
    window_inds : numpy.ndarray
        Vector of indices for windows.
        During training, batches of windows are made
        by grabbing indices randomly from this vector,
        then getting windows of the specified size
        from the arrays representing the input data
        and targets for the neural network.
    """
    return np.arange(stop=n_frames - (window_size - 1), step=stride)


class TrainDatapipe:
    """Dataset used for training neural network models
    on the frame classification task,
    where the source data consists of audio signals
    or spectrograms of varying lengths.

    Unlike
    :class:`vak.datasets.frame_classification.InferDatapipe`,
    this class does not return entire samples
    from the source dataset.
    Instead each paired samples :math:`(x_i, y_i)`
    returned by this dataset class consists of
    a window :math:`x_i` of fixed length
    :math:`w` from the underlying data ``X`` of total length :math:`T`.
    Each :math:`y_i` is a vector of the same size :math:`w`, containing
    an integer class label for each *frame* in the window :math:`x_i`.
    The entire dataset consists of some number of windows
    :math:`I` determined by a ``stride`` parameter :math:`s`,
    :math:`I = (T - w) / s`.

    The underlying data consists of single arrays
    for both the input to the network ``X``
    and the targets for the network output ``Y``.
    These single arrays ``X`` and ``Y`` are
    created by concatenating samples from the source
    data, e.g., audio files or spectrogram arrays.
    (This is true for
    :class:`vak.datasets.frame_classification.InferDatapipe`
    as well.)
    The dimensions of :math:`X`  will be (channels, ..., frames),
    i.e., audio will have dimensions (channels, samples)
    and spectrograms will have dimensions
    (channels, frequency bins, time bins).
    The signal :math:`X` may be either audio or spectrogram,
    meaning that a frame will be either a single sample
    in an audio signal or a single time bin in a spectrogram.
    The last dimension of ``X`` will always be the
    number of total frames in the dataset,
    either audio samples or spectrogram time bins,
    and ``Y`` will be the same size, containing
    an integer class label for each frame.

    Attributes
    ----------
    dataset_path : pathlib.Path
        Path to directory that represents a
        frame classification dataset,
        as created by
        :func:`vak.prep.prep_frame_classification_dataset`.
    split : str
        The name of a split from the dataset,
        one of {'train', 'val', 'test'}.
    subset : str, optional
        Name of subset to use.
        If specified, this takes precedence over split.
        Subsets are typically taken from the training data
        for use when generating a learning curve.
    dataset_df : pandas.DataFrame
        A frame classification dataset,
        represented as a :class:`pandas.DataFrame`.
        This will be only the rows that correspond
        to either ``subset`` or ``split`` from the
        ``dataset_df`` that was passed in when
        instantiating the class.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.
    frame_paths : numpy.ndarray
        Paths to npy files containing frames,
        either spectrograms or audio signals
        that are input to the model.
    frame_labels_paths : numpy.ndarray
        Paths to npy files containing vectors
        with a label for each frame.
        The targets for the outputs of the model.
    sample_ids : numpy.ndarray
        Indexing vector representing which sample
        from the dataset every frame belongs to.
    inds_in_sample : numpy.ndarray
        Indexing vector representing which index
        within each sample from the dataset
        that every frame belongs to.
    window_size : int
        Size of windows to return;
        number of frames.
    frames_standardizer : vak.transforms.FramesStandardizer, optional
        Transform applied to frames, the input to the neural network model.
        Optional, default is None.
        If supplied, will be used with the transform applied to inputs and targets,
        :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.
    frame_dur: float
        Duration of a frame, i.e., a single sample in audio
        or a single timebin in a spectrogram.
    stride : int
        The size of the stride used to determine which windows
        are included in the dataset. The default is 1.
        Used to compute ``window_inds``,
        with the function
        :func:`vak.datasets.frame_classification.train_datapipe.get_window_inds`.
    window_inds : numpy.ndarray, optional
        A vector of valid window indices for the dataset.
        If specified, this takes precedence over ``stride``.
    frames_standardizer : vak.transforms.FramesStandardizer, optional
        Transform applied to frames, the input to the neural network model.
        Optional, default is None.
        If supplied, will be used with the transform applied to inputs and targets,
        :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.
    """

    def __init__(
        self,
        dataset_path: str | pathlib.Path,
        dataset_df: pd.DataFrame,
        input_type: str,
        split: str,
        sample_ids: npt.NDArray,
        inds_in_sample: npt.NDArray,
        window_size: int,
        frame_dur: float,
        stride: int = 1,
        subset: str | None = None,
        window_inds: npt.NDArray | None = None,
        frames_standardizer: FramesStandardizer | None = None,
    ):
        """Initialize a new instance of a TrainDatapipe.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            frame classification dataset,
            as created by
            :func:`vak.prep.prep_frame_classification_dataset`.
        dataset_df : pandas.DataFrame
            A frame classification dataset,
            represented as a :class:`pandas.DataFrame`.
        input_type : str
            The type of input to the neural network model.
            One of {'audio', 'spect'}.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
        sample_ids : numpy.ndarray
            Indexing vector representing which sample
            from the dataset every frame belongs to.
        inds_in_sample : numpy.ndarray
            Indexing vector representing which index
            within each sample from the dataset
            that every frame belongs to.
        window_size : int
            Size of windows to return;
            number of frames.
        frame_dur: float
            Duration of a frame, i.e., a single sample in audio
            or a single timebin in a spectrogram.
        stride : int
            The size of the stride used to determine which windows
            are included in the dataset. The default is 1.
            Used to compute ``window_inds``,
            with the function
            :func:`vak.datasets.frame_classification.train_datapipe.get_window_inds`.
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.
        window_inds : numpy.ndarray, optional
            A vector of valid window indices for the dataset.
            If specified, this takes precedence over ``stride``.
        frames_standardizer : vak.transforms.FramesStandardizer, optional
            Transform applied to frames, the input to the neural network model.
            Optional, default is None.
            If supplied, will be used with the transform applied to inputs and targets,
            :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.
        """
        from ... import (
            prep,
        )  # avoid circular import, use for constants.INPUT_TYPES

        if input_type not in prep.constants.INPUT_TYPES:
            raise ValueError(
                f"``input_type`` must be one of: {prep.constants.INPUT_TYPES}\n"
                f"Value for ``input_type`` was: {input_type}"
            )

        self.dataset_path = pathlib.Path(dataset_path)
        self.split = split
        self.subset = subset
        # subset takes precedence over split, if specified
        if subset:
            dataset_df = dataset_df[dataset_df.subset == subset].copy()
        else:
            dataset_df = dataset_df[dataset_df.split == split].copy()
        self.dataset_df = dataset_df
        self.input_type = input_type
        self.frames_paths = self.dataset_df[
            constants.FRAMES_PATH_COL_NAME
        ].values
        self.frame_labels_paths = self.dataset_df[
            constants.MULTI_FRAME_LABELS_PATH_COL_NAME
        ].values
        self.sample_ids = sample_ids
        self.inds_in_sample = inds_in_sample
        self.window_size = window_size
        self.frame_dur = float(frame_dur)
        self.stride = stride
        if window_inds is None:
            window_inds = get_window_inds(
                sample_ids.shape[-1], window_size, stride
            )
        self.window_inds = window_inds
        self.item_transform = TrainItemTransform(
            frames_standardizer=frames_standardizer
        )

    @property
    def duration(self):
        return self.sample_ids.shape[-1] * self.frame_dur

    @property
    def shape(self):
        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        # used by vak functions that need to determine size of window,
        # e.g. when initializing a neural network model
        return tmp_item["frames"].shape

    def _load_frames(self, frames_path):
        """Helper function that loads "frames",
        the input to the frame classification model.
        Loads audio or spectrogram, depending on
        :attr:`self.input_type`.
        This function assumes that audio is in wav format
        and spectrograms are in npz files.
        """
        return helper.load_frames(frames_path, self.input_type)

    def __getitem__(self, idx):
        window_idx = self.window_inds[idx]
        sample_ids = self.sample_ids[
            window_idx : window_idx + self.window_size  # noqa: E203
        ]
        uniq_sample_ids = np.unique(sample_ids)
        if len(uniq_sample_ids) == 1:
            # we repeat ourselves here to avoid running a loop on one item
            sample_id = uniq_sample_ids[0]
            frames_path = self.dataset_path / self.frames_paths[sample_id]
            frames = self._load_frames(frames_path)
            frame_labels = np.load(
                self.dataset_path / self.frame_labels_paths[sample_id]
            )

        elif len(uniq_sample_ids) > 1:
            frames = []
            frame_labels = []
            for sample_id in sorted(uniq_sample_ids):
                frames_path = self.dataset_path / self.frames_paths[sample_id]
                frames.append(self._load_frames(frames_path))
                frame_labels.append(
                    np.load(
                        self.dataset_path / self.frame_labels_paths[sample_id]
                    )
                )

            if all([frames_.ndim == 1 for frames_ in frames]):
                # --> all 1-d audio vectors; if we specify `axis=1` here we'd get error
                frames = np.concatenate(frames)
            else:
                frames = np.concatenate(frames, axis=1)
            frame_labels = np.concatenate(frame_labels)
        else:
            raise ValueError(
                f"Unexpected number of ``uniq_sample_ids``: {uniq_sample_ids}"
            )

        inds_in_sample = self.inds_in_sample[window_idx]
        frames = frames[
            ...,
            inds_in_sample : inds_in_sample + self.window_size,  # noqa: E203
        ]
        frame_labels = frame_labels[
            inds_in_sample : inds_in_sample + self.window_size  # noqa: E203
        ]
        item = self.item_transform(frames, frame_labels)
        return item

    def __len__(self):
        """number of batches"""
        return len(self.window_inds)

    @classmethod
    def from_dataset_path(
        cls,
        dataset_path: str | pathlib.Path,
        window_size: int,
        stride: int = 1,
        split: str = "train",
        subset: str | None = None,
        frames_standardizer: FramesStandardizer | None = None,
    ):
        """Make a :class:`TrainDatapipe` instance,
        given the path to a frame classification dataset.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            frame classification dataset,
            as created by
            :func:`vak.prep.prep_frame_classification_dataset`.
        window_size : int
            Size of windows to return;
            number of frames.
        stride : int
            The size of the stride used to determine which windows
            are included in the dataset. The default is 1.
            Used to compute ``window_inds``,
            with the function
            :func:`vak.datasets.frame_classification.train_datapipe.get_window_inds`.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.
        frames_standardizer : vak.transforms.FramesStandardizer, optional
            Transform applied to frames, the input to the neural network model.
            Optional, default is None.
            If supplied, will be used with the transform applied to inputs and targets,
            :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.

        Returns
        -------
        dataset : vak.datasets.frame_classification.TrainDatapipe
        """
        dataset_path = pathlib.Path(dataset_path)
        metadata = Metadata.from_dataset_path(dataset_path)
        frame_dur = metadata.frame_dur
        input_type = metadata.input_type

        dataset_csv_path = dataset_path / metadata.dataset_csv_filename
        dataset_df = pd.read_csv(dataset_csv_path)

        split_path = dataset_path / split
        if subset:
            sample_ids_path = (
                split_path
                / helper.sample_ids_array_filename_for_subset(subset)
            )
        else:
            sample_ids_path = split_path / constants.SAMPLE_IDS_ARRAY_FILENAME
        sample_ids = np.load(sample_ids_path)

        if subset:
            inds_in_sample_path = (
                split_path
                / helper.inds_in_sample_array_filename_for_subset(subset)
            )
        else:
            inds_in_sample_path = (
                split_path / constants.INDS_IN_SAMPLE_ARRAY_FILENAME
            )
        inds_in_sample = np.load(inds_in_sample_path)

        window_inds_path = split_path / constants.WINDOW_INDS_ARRAY_FILENAME
        if window_inds_path.exists():
            window_inds = np.load(window_inds_path)
        else:
            window_inds = None

        return cls(
            dataset_path,
            dataset_df,
            input_type,
            split,
            sample_ids,
            inds_in_sample,
            window_size,
            frame_dur,
            stride,
            subset,
            window_inds,
            frames_standardizer,
        )



"""Tests for vak.predict.predict module."""
from unittest import mock
import pytest

import vak.cli.predict
import vak.config
import vak.common.constants
import vak.common.paths


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format, model_name, predict_function_to_mock",
    [
        ("cbin", None, "notmat", "TweetyNet",
         'vak.predict.predict_.predict_with_frame_classification_model'),
    ],
)
def test_predict(
    audio_format, spect_format, annot_format, model_name, predict_function_to_mock,
        specific_config_toml_path, tmp_path
):
    """Test that :func:`vak.predict.predict` dispatches to the correct model-specific
    training functions"""
    output_dir = tmp_path.joinpath(
        f"test_predict_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "predict", "key": "output_dir", "value": str(output_dir)},
        {"table": "predict", "key": "trainer", "value": {"accelerator": "cpu", "devices": 1}},
    ]

    toml_path = specific_config_toml_path(
        config_type="predict",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    results_path = tmp_path / 'results_path'
    results_path.mkdir()

    with mock.patch(predict_function_to_mock, autospec=True) as mock_predict_function:
        vak.predict.predict(
            model_config=cfg.predict.model.asdict(),
            dataset_config=cfg.predict.dataset.asdict(),
            trainer_config=cfg.predict.trainer.asdict(),
            checkpoint_path=cfg.predict.checkpoint_path,
            labelmap_path=cfg.predict.labelmap_path,
            num_workers=cfg.predict.num_workers,
            timebins_key=cfg.prep.spect_params.timebins_key,
            frames_standardizer_path=cfg.predict.frames_standardizer_path,
            annot_csv_filename=cfg.predict.annot_csv_filename,
            output_dir=cfg.predict.output_dir,
            min_segment_dur=cfg.predict.min_segment_dur,
            majority_vote=cfg.predict.majority_vote,
            save_net_outputs=cfg.predict.save_net_outputs,
        )
        assert mock_predict_function.called



"""Constants used by :mod:`vak.prep`.

Defined in a separate module to minimize circular imports.
"""

from . import frame_classification, parametric_umap

VALID_PURPOSES = frozenset(
    [
        "eval",
        "learncurve",
        "predict",
        "train",
    ]
)

INPUT_TYPES = {"audio", "spect"}

DATASET_TYPE_FUNCTION_MAP = {
    "frame classification": frame_classification.prep_frame_classification_dataset,
    "parametric umap": parametric_umap.prep_parametric_umap_dataset,
}

DATASET_TYPES = tuple(DATASET_TYPE_FUNCTION_MAP.keys())



"""Class that represents dataset table in configuration file."""

from __future__ import annotations

import pathlib

import attr.validators
from attr import asdict, define, field

from ..common.converters import expanded_user_path


@define
class DatasetConfig:
    """Class that represents dataset table in configuration file.

    Attributes
    ----------
    path : pathlib.Path
        Path to the directory that contains the dataset.
        Equivalent to the `root` parameter of :mod:`torchvision`
        datasets.
    splits_path : pathlib.Path, optional
        Path to file representing splits.
        Default is None.
    name : str, optional
        Name of dataset. Only required for built-in datasets
        from the :mod:`~vak.datasets` module. Default is None.
    params: dict, optional
        Parameters for dataset class,
        passed in as keyword arguments.
        E.g., ``window_size=2000``.
        Default is None.
    """

    path: pathlib.Path = field(converter=expanded_user_path)
    splits_path: pathlib.Path | None = field(
        converter=attr.converters.optional(expanded_user_path), default=None
    )
    name: str | None = field(
        converter=attr.converters.optional(str), default=None
    )
    params: dict | None = field(
        # we default to an empty dict instead of None
        # so we can still do **['dataset']['params'] everywhere we do when params are specified
        converter=attr.converters.optional(dict),
        default={},
    )

    @classmethod
    def from_config_dict(cls, config_dict: dict) -> DatasetConfig:

        return cls(**config_dict)

    def asdict(self):
        """Convert this :class:`DatasetConfig` instance
        to a :class:`dict` that can be passed
        into functions that take a ``dataset_config`` argument,
        like :func:`vak.train` and :func:`vak.predict`.
        """
        return asdict(self)



"""Code that handles classes that represent the definition
of a neural network model; the abstraction of how models
are declared with code in vak."""

from __future__ import annotations

import dataclasses
import inspect
from typing import Type, Union

import torch

REQUIRED_MODEL_DEFINITION_CLASS_VARS = (
    "network",
    "loss",
    "optimizer",
    "metrics",
    "default_config",
)

VALID_CONFIG_KEYS = REQUIRED_MODEL_DEFINITION_CLASS_VARS[
    :-1
]  # everything but 'default_config'


@dataclasses.dataclass
class ModelDefinition:
    """A class that represents the definition of a neural network model.

    A model definition is any class that has the following class variables:

    network: torch.nn.Module or dict
        Neural network.
        If a dict, should map string network names to torch.nn.Module classes.
    loss: torch.nn.Module, callable
        Either a built-in loss module, or a callable function that computes loss.
    optimizer: torch.optim.Optimizer
        Optimizer used to optimize neural network parameters during training.
    metrics: dict
        Metrics used to evaluate network. Should map string names of metric
        to callable classes that compute metric.
    default_config : dict
        That specifies default keyword arguments to use when instantiating any classes
        in ``network``, ``optimizer``, ``loss``, or ``metrics``.
        Used by ``vak.models.base.Model`` and its
        sub-classes that represent model families. E.g., those classes will do:
        ``network = self.definition.network(**self.definition.default_config['network'])``.

    Note it is **not** necessary to sub-class this class;
    it exists mainly for type-checking purposes.

    For more detail, see :func:`vak.models.decorator.model`
    and :class:`vak.models.ModelFactory`.
    """

    network: Union[torch.nn.Module, dict]
    loss: dict
    optimizer: torch.optim.Optimizer
    metrics: dict
    default_config: dict


# default that we set ``definition.default_config`` to,
# if definition does not have that class variable declared
DEFAULT_DEFAULT_CONFIG = {
    "network": {},
    "loss": {},
    "optimizer": {},
    "metrics": {},
}


def validate(definition: Type) -> Type:
    """Validate a model definition.

    A model definition is a class that has the following class variables:
        network: torch.nn.Module or dict
            Neural network.
            If a dict, should map string network names to torch.nn.Module classes.
        loss: torch.nn.Module, callable
            Either a built-in loss module, or a callable function that computes loss.
        optimizer: torch.optim.Optimizer
            Optimizer used to optimize neural network parameters during training.
        metrics: dict
            Metrics used to evaluate network. Should map string names of metric
            to callable classes that compute metric.
        default_config : dict
            That specifies default keyword arguments to use when instantiating any classes
            in ``network``, ``optimizer``, ``loss``, or ``metrics``.
            Used by ``vak.models.base.Model`` and its
            sub-classes that represent model families. E.g., those classes will do:
            ``network = self.definition.network(**self.definition.default_config['network'])``.
            If this class variable is not specified, it defaults to a ``dict``
            with the required keys, that map to empty ``dicts``.

    By providing this abstraction, ``vak`` commits in code
    to the idea that a neural network model consists of just
    the network function(s), the optimizer and the loss used
    to optimize the parameters of the network(s),
    as measured with the metrics.

    Parameters
    ----------
    definition : ModelDefinition
        A definition of a neural network model.
        A class having the class variables
        described above, with specific
        classes / callables / dicts assigned
        to those class variables.
        For an example, see
        ``vak.models.tweetynet.TweetyNet``.
        Does **not** need to be a sub-class of
        ``vak.models.definition.ModelDefinition``
        (that is used for type checking).

    Returns
    -------
    definition : type
        After validation, with ``default_config`` set to default
        if none was specified, as described above.

    Notes
    -----
    This function is used by the decorator
    ``vak.decorator.model``
    to validate a definition when
    converting it into a sub-class ofhttps://peps.python.org/pep-0416/
    ``vak.models.Model``.

    It's also used by :class:`vak.models.ModelFactory`,
    to validate a definition before building
    a new model instance from the definition.
    """
    # need to set this default first
    # so we don't throw error when checking class variables
    # if user did not specify ``default_config``
    if not hasattr(definition, "default_config"):
        definition.default_config = DEFAULT_DEFAULT_CONFIG
    else:
        # if they **did** specify ``default_config``,
        # make sure it's a dict
        if not isinstance(definition.default_config, dict):
            raise TypeError(
                "A model definition's ``default_config`` must be ``dict`` (or None)"
                f"but the type was: {type(definition.default_config)}"
            )

    # ---- check if any required class variables are missing
    definition_vars = {
        key: val
        for key, val in vars(definition).items()
        # keep class vars; throw out __module__, __doc__, etc.
        if not (key.startswith("__") and key.endswith("__"))
    }
    definition_class_var_names = list(definition_vars.keys())
    if not all(
        [
            expected_class_var_name in definition_class_var_names
            for expected_class_var_name in REQUIRED_MODEL_DEFINITION_CLASS_VARS
        ]
    ):
        missing_var_name = set(REQUIRED_MODEL_DEFINITION_CLASS_VARS) - set(
            definition_class_var_names
        )
        raise ValueError(
            f"Model definition is missing the following class variable(s): {missing_var_name}"
        )

    # ---- check if there are any extra class variables
    if any(
        [
            modeldef_var_name not in REQUIRED_MODEL_DEFINITION_CLASS_VARS
            for modeldef_var_name in definition_class_var_names
        ]
    ):
        extra_var_name = set(definition_class_var_names) - set(
            REQUIRED_MODEL_DEFINITION_CLASS_VARS
        )
        raise ValueError(
            f"Model definition has invalid class variable(s): {extra_var_name}."
            f"Valid class variables are: {REQUIRED_MODEL_DEFINITION_CLASS_VARS}"
        )

    # ---- now for each class variable check if they are the expected type.
    # either a torch.nn.Module or torch.optim.Optimizer subclass,
    # a dict mapping string names to torch.nn.Modules, or
    # a dict mapping string names to Callables.
    # Note that it's still hard to "unstringify" type annotations,
    # esp. in Python < 3.10, so
    # instead of getting it dynamically from __annotations__
    # we do validation "by hand" which is very verbose

    # ---- validate network
    network_obj = getattr(definition, "network")
    if inspect.isclass(network_obj):
        if not issubclass(network_obj, torch.nn.Module):
            raise TypeError(
                "A model definition's 'network' variable must be a subclass of torch.nn.Module "
                "or a dict mapping string names to torch.nn.Module subclasses, "
                f"but type was: {type(network_obj)}"
            )
    elif isinstance(network_obj, dict):
        for network_dict_key, network_dict_val in network_obj.items():
            if not isinstance(network_dict_key, str):
                raise TypeError(
                    "A model definition with a ``network`` variable that is a dict "
                    "should have keys that are strings, "
                    f"but the following key has type {type(network_dict_key)}: {network_dict_key}"
                )
            if not issubclass(network_dict_val, torch.nn.Module):
                raise TypeError(
                    "A model definition with a ``network`` variable that is a dict "
                    f"should have string keys mapping to values that are torch.nn.Module subclasses, "
                    f"but the following value has type {type(network_dict_val)}: {network_dict_val}"
                )
    else:
        raise TypeError(
            "A model definition's 'network' variable must be a subclass of torch.nn.Module "
            "or a dict mapping string names to torch.nn.Module subclasses, "
            f"but type was: {type(network_obj)}"
        )

    # ---- validate loss
    loss_obj = getattr(definition, "loss")
    # need complicated if-else here because issubclass throws an error if we don't pass it a class
    if inspect.isclass(loss_obj):
        if issubclass(loss_obj, torch.nn.Module):
            invalid_loss_obj_type = False
        else:
            invalid_loss_obj_type = True
    else:
        if inspect.isfunction(loss_obj):
            invalid_loss_obj_type = False
        else:
            invalid_loss_obj_type = True
    if invalid_loss_obj_type:
        raise TypeError(
            "A model definition's 'loss' variable must be a subclass of torch.nn.Module or a function, "
            f"but type was: {type(loss_obj)}"
        )

    # ---- validate optimizer
    optim_obj = getattr(definition, "optimizer")
    if not issubclass(optim_obj, torch.optim.Optimizer):
        raise TypeError(
            "A model definition's 'optimizer' variable must be a subclass of torch.optim.Optimizer, "
            f"but type was: {type(optim_obj)}"
        )

    # ---- validate metrics
    metrics_obj = getattr(definition, "metrics")
    if not isinstance(metrics_obj, dict):
        raise TypeError(
            "A model definition's 'metrics' variable must be a dict mapping string names to callables, "
            f"but was type: {type(metrics_obj)}"
        )
    for metrics_dict_key, metrics_dict_val in metrics_obj.items():
        if not isinstance(metrics_dict_key, str):
            raise TypeError(
                f"A model definition's 'metrics' variable must be a dict mapping string names to callables, "
                f"but the following key has type {type(metrics_dict_key)}: {metrics_dict_key}"
            )
        if not (
            inspect.isclass(metrics_dict_val) and callable(metrics_dict_val)
        ):
            raise TypeError(
                "A model definition's 'metrics' variable must be a dict mapping "
                "string names to classes that define __call__ methods, "
                f"but the key '{metrics_dict_key}' maps to a value with type {type(metrics_dict_val)}, "
                f"not recognized as callable."
            )

    # ---- validate default config
    default_config = getattr(definition, "default_config")

    if not all(
        [
            config_key in VALID_CONFIG_KEYS
            for config_key in default_config.keys()
        ]
    ):
        invalid_keys = [
            config_key
            for config_key in default_config.keys()
            if config_key not in VALID_CONFIG_KEYS
        ]
        raise ValueError(
            f"Invalid keys in default_config: {invalid_keys}."
            f"Valid keys are: {VALID_CONFIG_KEYS}"
        )

    # -------- validate 'network' config
    network_config = default_config.get("network")
    if network_config is None:
        if inspect.isclass(definition.network):
            # calling 'if issubclass(definition.network, torch.nn.Module)'
            # would raise an error when definition.network is a dict
            definition.default_config["network"] = {}
        elif isinstance(definition.network, dict):
            definition.default_config["network"] = {
                network_name: {} for network_name in definition.network.keys()
            }
    elif len(network_config) > 0:
        if inspect.isclass(definition.network):
            # calling 'if issubclass(definition.network, torch.nn.Module)'
            # would raise an error when definition.network is a dict
            network_init_params = list(
                inspect.signature(
                    definition.network.__init__
                ).parameters.keys()
            )
            if any(
                [
                    network_kwarg not in network_init_params
                    for network_kwarg in network_config.keys()
                ]
            ):
                invalid_keys = set(network_config.keys()) - set(
                    network_init_params
                )
                raise ValueError(
                    f"The following keyword arguments specified in the ``default_config`` "
                    f"for ``network`` are invalid: {invalid_keys}."
                    f"Valid arguments are: {network_init_params}"
                )

        elif isinstance(definition.network, dict):
            if any(
                [
                    network_name not in definition.network.keys()
                    for network_name in network_config.keys()
                ]
            ):
                invalid_network_names = [
                    network_name
                    for network_name in network_config.keys()
                    if network_name not in definition.network.keys()
                ]

                raise ValueError(
                    "When model definition's ``network`` is a ``dict`` mapping string names to ``torch.nn.Module``s,"
                    "the definition's ``default_config`` should have only those string names as keys."
                    f"The following keys in the default_config for network are invalid: {invalid_network_names}."
                    f"Valid keys are these network names: {definition.network.keys()}"
                    "Please rewrite ``default_config`` so keys of ``default_config['network']`` "
                    "are only those string names, "
                    "and the corresponding values for those keys are keyword arguments for the networks."
                )
            for network_name, network_kwargs in network_config.items():
                network_init_params = list(
                    inspect.signature(
                        definition.network[network_name].__init__
                    ).parameters.keys()
                )
                if any(
                    [
                        network_kwarg not in network_init_params
                        for network_kwarg in network_kwargs.keys()
                    ]
                ):
                    invalid_keys = set(network_config.keys()) - set(
                        network_init_params
                    )
                    raise ValueError(
                        f"The following keyword arguments specified in the ``default_config`` "
                        f"for ``network`` are invalid: {invalid_keys}."
                        f"Valid arguments are: {network_init_params}"
                    )

    # -------- validate 'loss' config
    loss_config = default_config.get("loss")
    if loss_config is None:
        definition.default_config["loss"] = {}
    elif len(loss_config) > 0:
        if inspect.isfunction(definition.loss):
            raise ValueError(
                "Model definition's default_config specifies keyword arguments for loss, "
                "but loss is a function, not a class. Please only specify keyword arguments for classes."
            )
        loss_init_params = list(
            inspect.signature(definition.loss.__init__).parameters.keys()
        )
        if any(
            [
                loss_kwarg not in loss_init_params
                for loss_kwarg in loss_config.keys()
            ]
        ):
            invalid_loss_kwargs = set(loss_config.keys()) - set(
                loss_init_params
            )
            raise ValueError(
                f"The following keyword arguments specified in the ``default_config`` "
                f"for ``loss`` are invalid: {invalid_loss_kwargs}."
                f"Valid arguments are: {loss_init_params}"
            )

    # -------- validate 'optimizer' config
    optimizer_config = default_config.get("optimizer")
    if optimizer_config is None:
        definition.default_config["optimizer"] = {}
    elif len(optimizer_config) > 0:
        optimizer_init_params = list(
            inspect.signature(definition.optimizer.__init__).parameters.keys()
        )
        if any(
            [
                optimizer_kwarg not in optimizer_init_params
                for optimizer_kwarg in optimizer_config.keys()
            ]
        ):
            invalid_optimizer_kwargs = set(optimizer_config.keys()) - set(
                optimizer_init_params
            )
            raise ValueError(
                f"The following keyword arguments specified in the ``default_config`` "
                f"for ``optimizer`` are invalid: {invalid_optimizer_kwargs}."
                f"Valid arguments are: {optimizer_init_params}"
            )

    # -------- validate 'metrics' config
    metrics_config = default_config.get("metrics")
    if metrics_config is None:
        definition.default_config["metrics"] = {}
    elif len(metrics_config) > 0:
        if any(
            [
                metric_name not in definition.metrics
                for metric_name in metrics_config.keys()
            ]
        ):
            invalid_metric_names = set(metrics_config.keys()) - set(
                definition.metrics.keys()
            )
            raise ValueError(
                f"The following metric names specified in the ``default_config`` "
                f"for ``metrics`` are invalid: {invalid_metric_names}."
                f"Valid metric names are: {definition.metrics.keys()}"
            )
        for metric_name, metric_class_config in metrics_config.items():
            metric_class_init_params = list(
                inspect.signature(
                    definition.metrics[metric_name].__init__
                ).parameters.keys()
            )
            if any(
                [
                    metric_class_kwarg not in metric_class_init_params
                    for metric_class_kwarg in metric_class_config.keys()
                ]
            ):
                invalid_metric_class_kwargs = set(
                    metric_class_config.keys()
                ) - set(metric_class_init_params)
                raise ValueError(
                    f"The following keyword arguments specified in the ``default_config`` "
                    f"for 'metrics' class {definition.metrics[metric_name]} are invalid: "
                    f"{invalid_metric_class_kwargs}."
                    f"Valid arguments are: {metric_class_init_params}"
                )

    return definition



from __future__ import annotations

import logging
import pathlib

import crowsetta
import dask.bag as db
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar

from ..common import annotation, constants
from ..common.converters import expanded_user_path, labelset_to_set
from ..common.typing import PathLike
from .spectrogram_dataset.audio_helper import files_from_dir

logger = logging.getLogger(__name__)


# constant, used for names of columns in DataFrame below
DF_COLUMNS = [
    "audio_path",
    "annot_path",
    "annot_format",
    "samplerate",
    "sample_dur",
    "duration",
]


def prep_audio_dataset(
    data_dir: PathLike,
    audio_format: str,
    annot_format: str | None = None,
    annot_file: str | pathlib.Path | None = None,
    labelset: set | None = None,
) -> pd.DataFrame:
    """Gets a set of audio files from a directory,
    optionally paired with an annotation file or files,
    and return a Pandas DataFrame that represents the set
    of files.

    Finds all files with ``audio_format`` in ``data_dir``,
    then finds any annotations with ``annot_format`` if specified,
    and additionally filter the audio and annotation files
    by ``labelset`` if specified.
    Then creates the dataframe with columns specified by
    ``vak.prep.audio_dataset.DF_COLUMNS``:
    ``"audio_path"``, ``"annot_path"``, ``"annot_format"``, ``"samplerate"``,
    ``"sample_dur",`` and ``"duration"``.

    Parameters
    ----------
    data_dir : str, pathlib.Path
        Path to directory containing audio files that should be used in dataset.
    audio_format : str
        A :class:`string` representing the format of audio files.
        One of :constant:`vak.common.constants.VALID_AUDIO_FORMATS`.
    annot_format : str
        Name of annotation format. Added as a column to the DataFrame if specified.
        Used by other functions that open annotation files via their paths from the DataFrame.
        Should be a format that the :mod:`crowsetta` library recognizes.
        Default is None.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    labelset : str, list, set
        Iterable of str or int, set of unique labels for annotations. Default is None.
        If not None, then files will be skipped where the associated annotation
        contains labels *not* found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using
        :func:`vak.common.converters.labelset_to_set`.
        See docstring of that function for details on how to specify ``labelset``.

    Returns
    -------
    source_files_df : pandas.Dataframe
        A set of source files that will be used to prepare a
        data set for use with neural network models,
        represented as a :class:`pandas.DataFrame`.
        Will contain paths to audio files,
        possibly paired with annotation files.
        The columns of the dataframe are specified by
        :const:`vak.prep.audio_dataset.DF_COLUMNS`.
    """
    # pre-conditions ---------------------------------------------------------------------------------------------------
    if audio_format not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"audio format must be one of '{constants.VALID_AUDIO_FORMATS}'; "
            f"format '{audio_format}' not recognized."
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    data_dir = expanded_user_path(data_dir)
    if not data_dir.exists() or not data_dir.is_dir():
        raise NotADirectoryError(
            f"`data_dir` not found, or not recognized as a directory:\n{data_dir}"
        )

    audio_files = files_from_dir(data_dir, audio_format)

    if annot_format is not None:
        if annot_file is None:
            annot_files = annotation.files_from_dir(
                annot_dir=data_dir, annot_format=annot_format
            )
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = [
                scribe.from_file(annot_file).to_annot()
                for annot_file in annot_files
            ]
        else:
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = scribe.from_file(annot_file).to_annot()
        if isinstance(annot_list, crowsetta.Annotation):
            # if e.g. only one annotated audio file in directory, wrap in a list to make iterable
            # fixes https://github.com/NickleDave/vak/issues/467
            annot_list = [annot_list]
    else:  # if annot_format not specified
        annot_list = None

    if annot_list:
        audio_annot_map = annotation.map_annotated_to_annot(
            audio_files, annot_list, annot_format
        )
    else:
        # no annotation, so map spectrogram files to None
        audio_annot_map = dict(
            (audio_path, None) for audio_path in audio_files
        )

    # use mapping (if generated/supplied) with labelset, if supplied, to filter
    if labelset:  # then remove annotations with labels not in labelset
        for audio_file, annot in list(audio_annot_map.items()):
            # loop in a verbose way (i.e. not a comprehension)
            # so we can give user warning when we skip files
            annot_labelset = set(annot.seq.labels)
            # below, set(labels_mapping) is a set of that dict's keys
            if not annot_labelset.issubset(set(labelset)):
                # because there's some label in labels that's not in labelset
                audio_annot_map.pop(audio_file)
                extra_labels = annot_labelset - labelset
                logger.info(
                    f"Found labels, {extra_labels}, in {pathlib.Path(audio_file).name}, "
                    "that are not in labels_mapping. Skipping file.",
                )

    # ---- actually make the dataframe ---------------------------------------------------------------------------------
    # this is defined here so all other arguments to 'to_dataframe' are in scope
    def _to_record(audio_annot_tuple):
        """helper function that enables parallelized creation of "records",
        i.e. rows for dataframe, from .
        Accepts a two-element tuple containing (1) a dictionary that represents a spectrogram
        and (2) annotation for that file"""
        audio_path, annot = audio_annot_tuple
        dat, samplerate = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](
            audio_path
        )
        sample_dur = 1.0 / samplerate
        audio_dur = dat.shape[-1] * sample_dur

        if annot is not None:
            annot_path = annot.annot_path
        else:
            annot_path = np.nan

        def abspath(a_path):
            if isinstance(a_path, str) or isinstance(a_path, pathlib.Path):
                return str(pathlib.Path(a_path).absolute())
            elif np.isnan(a_path):
                return a_path

        record = tuple(
            [
                abspath(audio_path),
                abspath(annot_path),
                (
                    annot_format
                    if annot_format
                    else constants.NO_ANNOTATION_FORMAT
                ),
                samplerate,
                sample_dur,
                audio_dur,
            ]
        )
        return record

    audio_path_annot_tuples = db.from_sequence(audio_annot_map.items())
    logger.info(
        "creating pandas.DataFrame representing dataset from audio files",
    )
    with ProgressBar():
        records = list(audio_path_annot_tuples.map(_to_record))

    return pd.DataFrame.from_records(data=records, columns=DF_COLUMNS)



import pytest

import vak.config.validators


def test_are_tables_valid(invalid_table_config_path):
    """test that invalid table name raises a ValueError"""
    config_dict = vak.config.load._load_toml_from_path(invalid_table_config_path)
    with pytest.raises(ValueError):
        vak.config.validators.are_tables_valid(
            config_dict, invalid_table_config_path
        )


def test_are_keys_valid(invalid_key_config_path):
    """test that table with an invalid key name raises a ValueError"""
    table_with_invalid_key = "prep"
    config_dict = vak.config.load._load_toml_from_path(invalid_key_config_path)
    with pytest.raises(ValueError):
        vak.config.validators.are_keys_valid(
            config_dict, table_with_invalid_key, invalid_key_config_path
        )



import numpy as np
import pytest

import vak
import vak.common  # for constants


from .test_functional import (
    FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
    MAX_ABS_DIFF,
    TIMEBIN_DUR_FOR_PARAMETRIZE,
    POSTPROCESS_PARAMS_ARGVALS,
    XFAIL_SPECT_FILES,
)


class TestFromSegments:
    def test_init(self):
        from_segments_tfm = vak.transforms.frame_labels.FromSegments()
        assert isinstance(from_segments_tfm, vak.transforms.frame_labels.FromSegments)

    @pytest.mark.parametrize(
        'annot, spect_path, labelset',
        FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
    )
    def test_call(self, annot, spect_path, labelset):
        labelset = vak.common.converters.labelset_to_set(labelset)
        labelmap = vak.common.labels.to_map(labelset, True)

        spect_dict = vak.common.files.spect.load(spect_path)
        timebins = spect_dict['t']

        try:
            lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
        except KeyError:
            pytest.skip(
                'Annotation with label not in labelset, would not include in dataset'
            )

        from_segments_tfm = vak.transforms.frame_labels.FromSegments(background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL])
        lbl_tb = from_segments_tfm(
            lbls_int,
            annot.seq.onsets_s,
            annot.seq.offsets_s,
            timebins,
        )
        assert lbl_tb.shape == timebins.shape
        assert all(
            [lbl in lbls_int for lbl in np.unique(lbls_int)]
        )


class TestToLabels:
    @pytest.mark.parametrize(
        'labelset',
        [tup[2] for tup in FROM_SEGMENTS_PARAMETRIZE_ARGVALS],
    )
    def test_init(self, labelset):
        # Note that we add an vak.common.constants.DEFAULT_BACKGROUND_LABEL class because post-processing transforms *require* it
        # This is default, just making it explicit
        labelset = vak.common.converters.labelset_to_set(labelset)
        labelmap = vak.common.labels.to_map(labelset, map_background=True)

        to_labels_tfm = vak.transforms.frame_labels.ToLabels(
            labelmap=labelmap,
        )
        assert isinstance(to_labels_tfm, vak.transforms.frame_labels.ToLabels)

    @pytest.mark.parametrize(
        "lbl_tb, labelmap, labels_expected_int",
        [
            (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, 'a': 1, 'b': 2}, [1, 2]),
            (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '1': 1, '2': 2}, [1, 2]),
            (np.array([0, 0, 21, 21, 0, 0, 22, 22, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '21': 21, '22': 22}, [21, 22]),
            (np.array([0, 0, 11, 11, 0, 0, 12, 12, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '11': 11, '12': 12}, [11, 12]),
        ]
    )
    def test_call(self, lbl_tb, labelmap, labels_expected_int):
        # Note that we add an vak.common.constants.DEFAULT_BACKGROUND_LABEL class because post-processing transforms *require* it
        # This is default, just making it explicit
        labelmap = vak.common.labels.multi_char_labels_to_single_char(
            labelmap, skip=(vak.common.constants.DEFAULT_BACKGROUND_LABEL,)
        )
        labelmap_inv = {v: k for k, v in labelmap.items()}
        labels_expected = ''.join([labelmap_inv[lbl_int] for lbl_int in labels_expected_int])

        to_labels_tfm = vak.transforms.frame_labels.ToLabels(
            labelmap=labelmap,
        )
        labels = to_labels_tfm(lbl_tb)
        assert labels == labels_expected

    @pytest.mark.parametrize(
        'annot, spect_path, labelset',
        FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
    )
    def test_call_real_data(
            self, annot, spect_path, labelset
    ):
        """test that ``to_labels_with_postprocessing`` recovers labels from real data"""
        labelset = vak.common.converters.labelset_to_set(labelset)
        labelmap = vak.common.labels.to_map(labelset)
        # next line, convert all labels to single characters
        # we can easily compare strings we get back with expected;
        # this is what core.eval does
        labelmap = vak.common.labels.multi_char_labels_to_single_char(
            labelmap, skip=(vak.common.constants.DEFAULT_BACKGROUND_LABEL,)
        )
        TIMEBINS_KEY = "t"

        if any(
            str(spect_path).endswith(spect_file_to_skip)
            for spect_file_to_skip in XFAIL_SPECT_FILES
        ):
            pytest.xfail(
                "Can't round trip segments -> lbl_tb -> segments "
                "because of small silent gap durations + large time bin durations"
            )

        try:
            lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
        except KeyError:
            pytest.skip(
                'Annotation with label not in labelset, would not include in dataset'
            )

        timebins = vak.common.files.spect.load(spect_path)[TIMEBINS_KEY]

        lbl_tb = vak.transforms.frame_labels.from_segments(
            lbls_int,
            annot.seq.onsets_s,
            annot.seq.offsets_s,
            timebins,
            background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
        )

        to_labels_tfm = vak.transforms.frame_labels.ToLabels(
            labelmap=labelmap,
        )
        labels = to_labels_tfm(lbl_tb)

        labelmap_multi_inv = {v: k for k, v in
                              labelmap.items()}
        labels_expected = "".join(
            [labelmap_multi_inv[lbl_int] for lbl_int in lbls_int]
        )
        assert labels == labels_expected


class TestToSegments:
    @pytest.mark.parametrize(
        'labelset',
        [tup[2] for tup in FROM_SEGMENTS_PARAMETRIZE_ARGVALS],
    )
    def test_init(self, labelset):
        # Note that we add an vak.common.constants.DEFAULT_BACKGROUND_LABEL class because post-processing transforms *require* it
        # This is default, just making it explicit
        labelset = vak.common.converters.labelset_to_set(labelset)
        labelmap = vak.common.labels.to_map(labelset, map_background=True)

        to_segments_tfm = vak.transforms.frame_labels.ToSegments(
            labelmap=labelmap,
        )
        assert isinstance(to_segments_tfm, vak.transforms.frame_labels.ToSegments)

    @pytest.mark.parametrize(
        'annot, spect_path, labelset',
        FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
    )
    def test_call_real_data(self, annot, spect_path, labelset):
        labelset = vak.common.converters.labelset_to_set(labelset)
        labelmap = vak.common.labels.to_map(labelset)

        TIMEBINS_KEY = "t"

        if any(
                str(spect_path).endswith(spect_file_to_skip)
                for spect_file_to_skip in XFAIL_SPECT_FILES
        ):
            pytest.xfail(
                "Can't round trip segments -> lbl_tb -> segments "
                "because of small silent gap durations + large time bin durations"
            )

        try:
            lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
        except KeyError:
            pytest.skip(
                'Annotation with label not in labelset, would not include in dataset'
            )

        timebins = vak.common.files.spect.load(spect_path)[TIMEBINS_KEY]

        lbl_tb = vak.transforms.frame_labels.from_segments(
            lbls_int,
            annot.seq.onsets_s,
            annot.seq.offsets_s,
            timebins,
            background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
        )

        to_segments_tfm = vak.transforms.frame_labels.ToSegments(
            labelmap=labelmap,
        )

        labels, onsets_s, offsets_s = to_segments_tfm(
            lbl_tb, timebins
        )

        assert np.all(np.char.equal(labels, annot.seq.labels))
        assert np.all(np.abs(annot.seq.onsets_s - onsets_s) < MAX_ABS_DIFF)
        assert np.all(np.abs(annot.seq.offsets_s - offsets_s) < MAX_ABS_DIFF)


class TestPostprocess:
    @pytest.mark.parametrize(
        'min_segment_dur, majority_vote, timebin_dur',
        # keep just the argvals we need to instantiate
        [argvals[3:5] + (TIMEBIN_DUR_FOR_PARAMETRIZE,) for argvals in POSTPROCESS_PARAMS_ARGVALS]
    )
    def test_init(self, min_segment_dur, majority_vote, timebin_dur):
        # Note that we add an vak.common.constants.DEFAULT_BACKGROUND_LABEL class
        # because post-processing transforms *require* it
        # This is default, just making it explicit
        to_labels_tfm = vak.transforms.frame_labels.PostProcess(
            min_segment_dur=min_segment_dur,
            majority_vote=majority_vote,
            timebin_dur=timebin_dur,
        )
        assert isinstance(to_labels_tfm, vak.transforms.frame_labels.PostProcess)

    @pytest.mark.parametrize(
        'lbl_tb, timebin_dur, background_label, min_segment_dur, majority_vote, lbl_tb_expected',
        POSTPROCESS_PARAMS_ARGVALS
    )
    def test_call(self, lbl_tb, timebin_dur, background_label, min_segment_dur, majority_vote, lbl_tb_expected):
        # Note that we add an vak.common.constants.DEFAULT_BACKGROUND_LABEL class because post-processing transforms *require* it
        # This is default, just making it explicit
        postprocess_tfm = vak.transforms.frame_labels.PostProcess(
            min_segment_dur=min_segment_dur,
            majority_vote=majority_vote,
            timebin_dur=timebin_dur,
        )

        lbl_tb = postprocess_tfm(
            lbl_tb
        )

        assert np.all(np.equal(lbl_tb, lbl_tb_expected))




"""Tests for vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset"""
import json
import pathlib
import shutil

import pandas as pd
from pandas.testing import assert_series_equal
import pytest

import vak


def assert_prep_output_matches_expected(dataset_path, df_returned_by_prep):
    dataset_path = pathlib.Path(dataset_path)
    assert dataset_path.exists()
    assert dataset_path.is_dir()

    log_path = sorted(dataset_path.glob('*log'))
    assert len(log_path) == 1

    meta_json_path = dataset_path / vak.datapipes.frame_classification.Metadata.METADATA_JSON_FILENAME
    assert meta_json_path.exists()

    with meta_json_path.open('r') as fp:
        meta_json = json.load(fp)

    dataset_csv_path = dataset_path / meta_json['dataset_csv_filename']
    assert dataset_csv_path.exists()

    df_from_dataset_path = pd.read_csv(dataset_csv_path)

    for column in vak.prep.spectrogram_dataset.spect_helper.DF_COLUMNS:
        if column == "duration":
            check_exact = False
        else:
            check_exact = True
        assert_series_equal(
            df_from_dataset_path[column],
            df_returned_by_prep[column],
            check_exact=check_exact,
        )

    if vak.datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME in df_returned_by_prep.columns:
        frames_paths = df_returned_by_prep[
            vak.datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME
        ].values
        for frames_path in frames_paths:
            assert (dataset_path / frames_path).exists()

    if vak.datapipes.frame_classification.constants.MULTI_FRAME_LABELS_PATH_COL_NAME in df_returned_by_prep.columns:
        frame_labels_paths = df_returned_by_prep[
            vak.datapipes.frame_classification.constants.MULTI_FRAME_LABELS_PATH_COL_NAME
        ].values
        if not all([frame_labels_path is None for frame_labels_path in frame_labels_paths]):
            for frame_labels_path in frame_labels_paths:
                assert (dataset_path / frame_labels_path).exists()


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("predict", "cbin", None, "notmat"),
        ("train", "cbin", None, "notmat"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_prep_frame_classification_dataset(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    specific_config_toml_path,
    default_model,
    tmp_path,
):
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {
            "table": "prep",
            "key": "output_dir",
            "value": str(output_dir),
        },
    ]
    toml_path = specific_config_toml_path(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = config_type.lower()
    dataset_df, dataset_path = vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
        data_dir=cfg.prep.data_dir,
        input_type=cfg.prep.input_type,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_params=cfg.prep.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        train_set_durs=cfg.prep.train_set_durs,
        num_replicates=cfg.prep.num_replicates,
    )

    assert_prep_output_matches_expected(dataset_path, dataset_df)


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("train", "cbin", None, "notmat"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_prep_frame_classification_dataset_raises_when_labelset_required_but_is_none(
    config_type,
    audio_format,
    spect_format,
    annot_format,
        specific_config_toml_path,
    default_model,
    tmp_path,
):
    """Test that `prep` raises a ValueError when the config
    requires a `labelset`,
    i.e., is one of {'train','learncurve', 'eval'},
    but it is left as None.

    Regression test for https://github.com/vocalpy/vak/issues/468.
    """
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "prep",
         "key": "output_dir",
         "value": str(output_dir),
         },
        {"table": "prep",
         "key": "labelset",
         "value": "DELETE-KEY",
         },
    ]
    toml_path = specific_config_toml_path(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = config_type.lower()
    with pytest.raises(ValueError):
        vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            purpose=purpose,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            output_dir=cfg.prep.output_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
        )


def test_prep_frame_classification_dataset_with_single_audio_and_annot(source_test_data_root,
                                                                       specific_config_toml_path,
                                                                       default_model,
                                                                       tmp_path):
    """
    regression test, checks that we avoid a repeat of
    https://github.com/vocalpy/vak/issues/467
    """
    data_dir = tmp_path / 'data_dir_with_single_audio_and_annot'
    data_dir.mkdir()
    source_data_dir = source_test_data_root / 'audio_cbin_annot_notmat/gy6or6/032412'
    cbins = sorted(source_data_dir.glob('*.cbin'))
    a_cbin = cbins[0]
    shutil.copy(a_cbin, data_dir)
    a_rec = a_cbin.parent / (a_cbin.stem + '.rec')
    assert a_rec.exists()
    shutil.copy(a_rec, data_dir)
    a_notmat = a_cbin.parent / (a_cbin.name + '.not.mat')
    assert a_notmat.exists()
    shutil.copy(a_notmat, data_dir)

    output_dir = tmp_path.joinpath(
        f"test_prep_eval_single_audio_and_annot"
    )
    output_dir.mkdir()

    keys_to_change = [
        {
            "table": "prep",
            "key": "data_dir",
            "value": str(data_dir),
        },
        {
            "table": "prep",
            "key": "output_dir",
            "value": str(output_dir),
        },
    ]

    toml_path = specific_config_toml_path(
        config_type='eval',
        model=default_model,
        audio_format='cbin',
        annot_format='notmat',
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = 'eval'
    dataset_df, dataset_path = vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
        data_dir=cfg.prep.data_dir,
        input_type=cfg.prep.input_type,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_params=cfg.prep.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
    )

    assert len(dataset_df) == 1


def test_prep_frame_classification_dataset_when_annot_has_single_segment(source_test_data_root,
                                                                         specific_config_toml_path,
                                                                         default_model,
                                                                         tmp_path):
    """
    regression test, checks that we avoid a repeat of
    https://github.com/vocalpy/vak/issues/466
    """
    data_dir = source_test_data_root / 'audio_cbin_annot_notmat' / 'gy6or6-song-edited-to-have-single-segment'

    output_dir = tmp_path.joinpath(
        f"test_prep_eval_annot_with_single_segment"
    )
    output_dir.mkdir()

    keys_to_change = [
        {
            "table": "prep",
            "key": "data_dir",
            "value": str(data_dir),
        },
        {
            "table": "prep",
            "key": "output_dir",
            "value": str(output_dir),
        },
    ]

    toml_path = specific_config_toml_path(
        config_type='eval',
        model=default_model,
        audio_format='cbin',
        annot_format='notmat',
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = 'eval'
    dataset_df, dataset_path = vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
        data_dir=cfg.prep.data_dir,
        input_type=cfg.prep.input_type,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_params=cfg.prep.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
    )

    assert len(dataset_df) == 1


@pytest.mark.parametrize(
    "dir_option_to_change",
    [
        {"table": "prep", "key": "data_dir", "value": '/obviously/does/not/exist/data'},
        {"table": "prep", "key": "output_dir", "value": '/obviously/does/not/exist/output'},
    ],
)
def test_prep_frame_classification_dataset_raises_not_a_directory(
    dir_option_to_change,
        specific_config_toml_path,
    default_model,
    tmp_path,
):
    """Test that `core.prep` raise NotADirectory error
    when one of the following is not a directory:
    data_dir, output_dir
    """
    toml_path = specific_config_toml_path(
        config_type="train",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=dir_option_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = "train"
    with pytest.raises(NotADirectoryError):
        vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            purpose=purpose,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            output_dir=cfg.prep.output_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
        )


@pytest.mark.parametrize(
    "path_option_to_change",
    [
        {"table": "prep", "key": "annot_file", "value": '/obviously/does/not/exist/annot.mat'},
    ],
)
def test_prep_frame_classification_dataset_raises_file_not_found(
    path_option_to_change,
        specific_config_toml_path,
    default_model,
    tmp_path,
):
    """Test that `core.prep` raise FileNotFound error
    when one of the following does not exist:
    annot_file

    Structuring unit test this way in case other path
    parameters get added.
    """
    toml_path = specific_config_toml_path(
        config_type="train",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=path_option_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    purpose = "train"
    with pytest.raises(FileNotFoundError):
        vak.prep.frame_classification.frame_classification.prep_frame_classification_dataset(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            purpose=purpose,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            output_dir=cfg.prep.output_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
        )



"""CLI parser for generate test data script"""
import argparse

from . import constants


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--step',
        choices=constants.GENERATE_TEST_DATA_STEPS,
        help=f"Which step of generating test data to perform, one of: {constants.GENERATE_TEST_DATA_STEPS}",
        default='all'
    )
    parser.add_argument(
        '--commands',
        choices=('train', 'learncurve', 'eval', 'predict', 'train_continue'),
        help=f"Space-separated list of commands to run for 'results' step",
        nargs="+",
        default=('train', 'learncurve', 'eval', 'predict', 'train_continue')
    )
    parser.add_argument(
        '--single-train-result',
        action = argparse.BooleanOptionalAction,
        help=(
            "If --single-train-result, require there be a single results directory "
            "from any training config when looking for them to use in other configs. "
            "If --no-single-train-result, allow multiple and use the most recent."
        )
    )
    return parser



"""Helper function that gets instances of classes representing datasets built into :mod:`vak`."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Mapping

from .. import common

Dataset = Mapping

if TYPE_CHECKING:
    from ..transforms import FramesStandardizer


def get(
    dataset_config: dict,
    split: Literal["predict", "test", "train", "val"],
    frames_standardizer: FramesStandardizer | None = None,
) -> Dataset:
    """Get an instance of a dataset class from :mod:`vak.datasets`.

    Parameters
    ----------
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    split : str
        Name of split to use.
        One of {"predict", "test", "train", "val"}.

    Returns
    -------
    dataset : class
        An instance of a class from :mod:`vak.datasets`,
        e.g. :class:`vak.datasets.BioSoundSegBench`.
    """
    from . import DATASETS  # avoid circular import

    if "name" not in dataset_config:
        raise KeyError(
            "A name is required to get a dataset, but "
            "`vak.datasets.get` received a `dataset_config` "
            f'without a "name":\n{dataset_config}'
        )
    if split not in common.constants.VALID_SPLITS:
        raise ValueError(
            f"Invalid value for `split`: {split}.\n"
            f"Valid splits are: {common.constants.VALID_SPLITS}"
        )

    dataset_name = dataset_config["name"]
    try:
        dataset_class = DATASETS[dataset_name]
    except KeyError as e:
        raise ValueError(
            f"Invalid dataset name: {dataset_name}\n."
            f"Built-in dataset names are: {DATASETS.keys()}"
        ) from e
    if frames_standardizer is not None:
        dataset_config["params"]["frames_standardizer"] = frames_standardizer
    dataset = dataset_class(
        dataset_path=dataset_config["path"],
        splits_path=dataset_config["splits_path"],
        split=split,
        **dataset_config["params"],
    )
    return dataset



"""Functional forms of transformations
related to frame labels,
i.e., vectors where each element represents
a label for a frame, either a single sample in audio
or a single time bin from a spectrogram.

This module is structured as followed:
- from_segments: transform to get frame labels from annotations
- to_labels: transform to get back just string labels from frame labels,
  used to evaluate a model
- to_segments: transform to get back segment onsets, offsets, and labels from frame labels.
  Inverse of ``from_segments``.
- post-processing transforms that can be used to "clean up" a vector of frame labels
  - segment_inds_list_from_class_labels: helper function used to find segments in a vector of frame labels
  - remove_short_segments: remove any segment less than a minimum duration
  - take_majority_vote: take a "majority vote" within each segment bounded by the background label,
    and apply the most "popular" label within each segment to all timebins in that segment
  - postprocess: combines remove_short_segments and take_majority_vote in one transform
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.stats

from ... import common
from ...common.timebins import timebin_dur_from_vec
from ...common.validators import column_or_1d, row_or_1d

__all__ = [
    # keep alphabetized
    "boundary_inds_from_boundary_labels",
    "from_segments",
    "postprocess",
    "remove_short_segments",
    "take_majority_vote",
    "segment_inds_list_from_class_labels",
    "to_labels",
    "to_segments",
]


def from_segments(
    labels_int: npt.NDArray,
    onsets_s: npt.NDArray,
    offsets_s: npt.NDArray,
    time_bins: npt.NDArray,
    background_label: int = 0,
) -> npt.NDArray:
    """Make a vector of labels for a vector of frames,
    given labeled segments in the form of onset times,
    offset times, and segment labels.

    Parameters
    ----------
    labels_int : list, numpy.ndarray
        A list or array of labels from the annotation for a vocalization,
        mapped to integers
    onsets_s : numpy.ndarray
        1-d vector of floats, segment onsets in seconds.
    offsets_s : numpy.ndarray
        1-d vector of floats, segment offsets in seconds.
    time_bins : numpy.ndarray
        1-d vector of floats, time in seconds for center of each time bin of a spectrogram.
    background_label : int
        Label assigned to frames that do not have labels associated with them.
        Default is 0.

    Returns
    -------
    frame_labels : numpy.ndarray
        same length as time_bins, with each element a label for each time bin
    """
    if (
        isinstance(labels_int, list)
        and not all([isinstance(lbl, int) for lbl in labels_int])
    ) or (
        isinstance(labels_int, np.ndarray)
        and labels_int.dtype not in [np.int8, np.int16, np.int32, np.int64]
    ):
        raise TypeError(
            "labels_int must be a list or numpy.ndarray of integers"
        )

    label_vec = (
        np.ones((time_bins.shape[-1],), dtype="int8") * background_label
    )
    onset_inds = [np.argmin(np.abs(time_bins - onset)) for onset in onsets_s]
    offset_inds = [
        np.argmin(np.abs(time_bins - offset)) for offset in offsets_s
    ]
    for label, onset, offset in zip(labels_int, onset_inds, offset_inds):
        # offset_inds[ind]+1 because offset time bin is still "part of" syllable
        label_vec[onset : offset + 1] = label  # noqa: E203

    return label_vec


def to_labels(
    frame_labels: npt.NDArray,
    labelmap: dict,
    background_label: str = common.constants.DEFAULT_BACKGROUND_LABEL,
) -> str:
    """Convert vector of frame labels to a string,
    one character for each continuous segment.

    Allows for converting output of network
    from a label for each frame
    to one label for each continuous segment,
    in order to compute string-based metrics like edit distance.

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        Typically, the output of a neural network.
    labelmap : dict
        That maps string labels to integers.
        The mapping is inverted to convert back to string labels.
    background_label: str, optional
        The string label applied to segments belonging to the
        background class.
        Default is
        :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.

    Returns
    -------
    labels : str
        The label at the onset of each continuous segment
        in ``frame_labels``, mapped back to string labels in ``labelmap``.
    """
    frame_labels = row_or_1d(frame_labels)

    onset_inds = np.diff(frame_labels, axis=0).astype(bool)
    onset_inds = np.insert(onset_inds, 0, True)

    labels = frame_labels[onset_inds]

    # remove background label
    if background_label in labelmap:
        labels = labels[labels != labelmap[background_label]]

    if len(labels) < 1:  # if removing all the 'unlabeled' leaves nothing
        return ""

    # only invert mapping and then map integer labels to characters
    inverse_labelmap = dict((v, k) for k, v in labelmap.items())
    labels = labels.tolist()
    labels = [inverse_labelmap[label] for label in labels]

    return "".join(labels)


def to_segments(
    frame_labels: npt.NDArray,
    labelmap: dict,
    frame_times: npt.NDArray,
    n_decimals_trunc: int = 5,
    background_label: str = common.constants.DEFAULT_BACKGROUND_LABEL,
) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
    """Convert a vector of frame labels
    into segments in the form of onset indices,
    offset indices, and labels.

    Finds where continuous runs of a single label start
    and stop in timebins, and considers each of these runs
    a segment.

    The function returns vectors of labels and onsets and offsets
    in units of seconds.

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        Output of a neural network.
    labelmap : dict
        That maps labels to consecutive integers.
        The mapping is inverted to convert back to labels.
    frame_times : numpy.ndarray
        Vector of times; the times are either the time of samples in audio,
        or the bin centers of columns in a spectrogram,
        returned by function that generated spectrogram.
        Used to convert onset and offset indices in frame_labels to seconds.
    n_decimals_trunc : int
        Number of decimal places to keep when truncating the timebin duration
        calculated from the vector of times t. Default is 5.

    Returns
    -------
    labels : numpy.ndarray
        Vector where each element is a label for a segment with its onset
        and offset indices given by the corresponding element in onset_inds
        and offset_inds.
    onsets_s : numpy.ndarray
        Vector where each element is the onset in seconds a segment.
        Each onset corresponds to the value at the same index in labels.
    offsets_s : numpy.ndarray
        Vector where each element is the offset in seconds of a segment.
        Each offset corresponds to the value at the same index in labels.
    """
    frame_labels = column_or_1d(frame_labels)

    if background_label in labelmap:
        # handle the case when all time bins are predicted to be unlabeled
        # see https://github.com/NickleDave/vak/issues/383
        uniq_frame_labels = np.unique(frame_labels)
        if (
            len(uniq_frame_labels) == 1
            and uniq_frame_labels[0] == labelmap[background_label]
        ):
            return None, None, None

    # used to find onsets/offsets below; compute here so if we fail we do so early
    timebin_dur = timebin_dur_from_vec(frame_times, n_decimals_trunc)

    offset_inds = np.nonzero(np.diff(frame_labels, axis=0))[
        0
    ]  # [0] because nonzero return tuple
    onset_inds = offset_inds + 1
    offset_inds = np.concatenate(
        (offset_inds, np.asarray([frame_labels.shape[0] - 1]))
    )
    onset_inds = np.concatenate((np.asarray([0]), onset_inds))
    labels = frame_labels[onset_inds]

    # remove background label
    if background_label in labelmap:
        keep = np.where(labels != labelmap[background_label])[0]
        labels = labels[keep]
        onset_inds = onset_inds[keep]
        offset_inds = offset_inds[keep]

    # handle case where removing 'unlabeled' leaves no segments
    if all([len(vec) == 0 for vec in (labels, onset_inds, offset_inds)]):
        return None, None, None

    inverse_labelmap = dict((v, k) for k, v in labelmap.items())
    labels = labels.tolist()
    labels = np.asarray([inverse_labelmap[label] for label in labels])
    # the 'best' estimate we can get of onset and offset times,
    # given binned times, and labels applied to each time bin,
    # is "some time" between the last labeled bin for one segment,
    # i.e. its offset, and the first labeled bin for the next
    # segment, i.e. its onset. In other words if the whole bin is labeled
    # as belonging to that segment, and the bin preceding it is labeled as
    # belonging to the previous section, then the onset of the current
    # segment must be the time between the two bins. To find those times
    # we use the bin centers and either subtract (for onsets) or add
    # (for offsets) half a timebin duration. This half a timebin
    # duration puts our onsets and offsets at the time "between" bins.
    onsets_s = frame_times[onset_inds] - (timebin_dur / 2)
    offsets_s = frame_times[offset_inds] + (timebin_dur / 2)

    # but this estimate will be "wrong" if we set the onset or offset time
    # outside the possible times in our timebin vector. Need to clean up.
    if onsets_s[0] < 0.0:
        onsets_s[0] = 0.0
    if offsets_s[-1] > frame_times[-1]:
        offsets_s[-1] = frame_times[-1]

    return labels, onsets_s, offsets_s


def segment_inds_list_from_class_labels(
    frame_labels: npt.NDArray, background_label: int = 0
) -> list[npt.NDArray]:
    """Given a vector of frame labels,
    returns a list of indexing vectors,
    one for each segment in the vector
    that is not labeled with the background label.

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
    background_label : int
        Label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.

    Returns
    -------
    segment_inds_list : list
        Of fancy indexing arrays. Each array can be used to index
        one segment in ``frame_labels``.
    """
    segment_inds = np.nonzero(frame_labels != background_label)[0]
    return np.split(segment_inds, np.where(np.diff(segment_inds) != 1)[0] + 1)


def remove_short_segments(
    frame_labels: npt.NDArray,
    segment_inds_list: list[npt.NDArray],
    timebin_dur: float,
    min_segment_dur: float | int,
    background_label: int = 0,
) -> tuple[npt.NDArray, list[npt.NDArray]]:
    """Remove segments from vector of frame labels
    that are shorter than a specified duration.

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        Output of a neural network.
    segment_inds_list : list
        Of numpy.ndarray, indices that will recover segments list from ``frame_labels``.
        Returned by function ``vak.labels.frame_labels_segment_inds_list``.
    timebin_dur : float
        Duration of a single timebin in the spectrogram, in seconds.
        Used to convert onset and offset indices in ``frame_labels`` to seconds.
    min_segment_dur : float
        Minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from frame_labels. Default is None, in which case no
        segments are removed.
    background_label : int
        Label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.

    Returns
    -------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        With segments whose duration is shorter than ``min_segment_dur``
        set to ``background_label``
    segment_inds_list : list
        Of numpy.ndarray, with arrays removed that represented
        segments in ``frame_labels`` that were shorter than ``min_segment_dur``.
    """
    new_segment_inds_list = []

    for segment_inds in segment_inds_list:
        if segment_inds.shape[-1] * timebin_dur < min_segment_dur:
            frame_labels[segment_inds] = background_label
            # DO NOT keep segment_inds array
        else:
            # do keep segment_inds array, don't change frame_labels
            new_segment_inds_list.append(segment_inds)

    return frame_labels, new_segment_inds_list


def take_majority_vote(
    frame_labels: npt.NDArray, segment_inds_list: list[npt.NDArray]
) -> npt.NDArray:
    """Transform segments containing multiple labels
    into segments with a single label by taking a "majority vote",
    i.e. assign all frames in the segment the most frequently
    occurring label in the segment.

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        Output of a neural network.
    segment_inds_list : list
        Of numpy.ndarray, indices that will recover segments list from frame_labels.
        Returned by function ``vak.labels.frame_labels_segment_inds_list``.

    Returns
    -------
    frame_labels : numpy.ndarray
        After the majority vote transform has been applied.
    """
    for segment_inds in segment_inds_list:
        segment = frame_labels[segment_inds]
        majority = scipy.stats.mode(segment, keepdims=False)[0].item()
        frame_labels[segment_inds] = majority

    return frame_labels


def boundary_inds_from_boundary_labels(
    boundary_labels: npt.NDArray, force_boundary_first_ind: bool = True
) -> npt.NDArray:
    """Return a :class:`numpy.ndarray` with the indices
    of boundaries, given a 1-D vector of boundary labels.

    Parameters
    ----------
    boundary_labels : numpy.ndarray
        Vector of integers ``{0, 1}``, where ``1`` indicates a boundary,
        and ``0`` indicates no boundary.
        Output of a frame classification model trained to classify each
        frame as "boundary" or "no boundary".
    force_boundary_first_ind : bool
        If ``True``, and the first index of ``boundary_labels`` is not classified as a boundary,
        force it to be a boundary.
    """
    boundary_labels = row_or_1d(boundary_labels)
    boundary_inds = np.nonzero(boundary_labels)[0]

    if force_boundary_first_ind:
        if len(boundary_inds) == 0:
            # handle edge case where no boundaries were predicted
            boundary_inds = np.array([0])  # replace with a single boundary, at index 0
        else:
            if boundary_inds[0] != 0:
                # force there to be a boundary at index 0
                np.insert(boundary_inds, 0, 0)

    return boundary_inds


def segment_inds_list_from_boundary_labels(
    boundary_labels: npt.NDArray, force_boundary_first_ind: bool = True
) -> list[npt.NDArray]:
    """Given an array of boundary labels,
    return a list of :class:`numpy.ndarray` vectors,
    each of which can be used to index one segment
    in a vector of frame labels.

    Parameters
    ----------
    boundary_labels : numpy.ndarray
        Vector of integers ``{0, 1}``, where ``1`` indicates a boundary,
        and ``0`` indicates no boundary.
        Output of a frame classification model trained to classify each
        frame as "boundary" or "no boundary".
    force_boundary_first_ind : bool
        If ``True``, and the first index of ``boundary_labels`` is not classified as a boundary,
        force it to be a boundary.

    Returns
    -------
    segment_inds_list : list
        Of fancy indexing arrays. Each array can be used to index
        one segment in ``frame_labels``.
    """
    boundary_inds = boundary_inds_from_boundary_labels(
        boundary_labels, force_boundary_first_ind
    )

    # at the end of `boundary_inds``, insert an imaginary "last" boundary we use just with ``np.arange`` below
    np.insert(boundary_inds, boundary_inds.shape[0], boundary_labels.shape[0])

    segment_inds_list = []
    for start, stop in zip(boundary_inds[:-1], boundary_inds[1:]):
        segment_inds_list.append(np.arange(start, stop))

    return segment_inds_list


def postprocess(
    frame_labels: npt.NDArray,
    timebin_dur: float,
    background_label: int = 0,
    min_segment_dur: float | None = None,
    majority_vote: bool = False,
    boundary_labels: npt.NDArray | None = None,
) -> npt.NDArray:
    """Apply post-processing transformations
    to a vector of frame labels.

    Optional post-processing
    consist of two transforms,
    that both rely on there being a label
    that corresponds to the background class.
    The first removes any segments that are
    shorter than a specified duration,
    by converting labels in those segments to the
    background class label.
    The second performs a "majority vote"
    transform within run of labels that is
    bordered on both sides by the "background" label.
    I.e., it counts the number of times any
    label occurs in that segment,
    and then assigns all bins the most common label.

    The function performs those steps in this order
    (pseudo-code):

    .. code-block::

       if min_segment_dur:
           frame_labels = remove_short_segments(frame_labels, labelmap, min_segment_dur)
       if majority_vote:
           frame_labels = majority_vote(frame_labels, labelmap)
       return frame_labels

    Parameters
    ----------
    frame_labels : numpy.ndarray
        A vector where each element represents
        a label for a frame, either a single sample in audio
        or a single time bin from a spectrogram.
        Output of a neural network.
    timebin_dur : float
        Duration of a time bin in a spectrogram,
        e.g., as estimated from vector of times
        using ``vak.timebins.timebin_dur_from_vec``.
    background_label : int
        Label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.
    min_segment_dur : float
        Minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from frame_labels. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        If True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment.
        This transform requires either a background label
        or a vector of boundary labels.
        Default is False.
    boundary_labels : numpy.ndarray, optional.
        Vector of integers ``{0, 1}``, where ``1`` indicates a boundary,
        and ``0`` indicates no boundary.
        Output of one head of a frame classification model,
        that has been trained to classify each frame as either
        "boundary" or "no boundary".
        Optional, default is None.
        If supplied, this vector is used to find segments
        before applying post-processing, instead
        of recovering them from ``frame_labels`` using the
        ``background_label``

    Returns
    -------
    frame_labels : numpy.ndarray
        Vector of frame labels after post-processing is applied.
    """
    frame_labels = row_or_1d(frame_labels)
    if boundary_labels is not None:
        boundary_labels = row_or_1d(boundary_labels)

    # handle the case when all time bins are predicted to be unlabeled
    # see https://github.com/NickleDave/vak/issues/383
    uniq_frame_labels = np.unique(frame_labels)
    if (
        len(uniq_frame_labels) == 1
        and uniq_frame_labels[0] == background_label
    ):
        return frame_labels  # -> no need to do any of the post-processing

    if boundary_labels is not None:
        segment_inds_list = segment_inds_list_from_boundary_labels(
            boundary_labels
        )
    else:
        segment_inds_list = segment_inds_list_from_class_labels(
            frame_labels, background_label=background_label
        )

    if min_segment_dur is not None:
        frame_labels, segment_inds_list = remove_short_segments(
            frame_labels,
            segment_inds_list,
            timebin_dur,
            min_segment_dur,
            background_label,
        )
        if len(segment_inds_list) == 0:  # no segments left after removing
            return frame_labels  # -> no need to do any of the post-processing

    if majority_vote:
        frame_labels = take_majority_vote(frame_labels, segment_inds_list)

    return frame_labels



import itertools

import pytest

import vak


LABELSETS = []
for labelset in (
    'abcde',
    '12345',
    [1, 2, 3, 4, 5],
    'mo', 'po', 'ta',
):
    LABELSETS.append(
        vak.common.converters.labelset_to_set(labelset)
    )
LABELMAPS = []
for labelset in LABELSETS:
    LABELMAPS.append(
        vak.common.labels.to_map(labelset, map_background=True)
    )

INPUT_SHAPES = (
    (1, 513, 88),
    (1, 267, 1000)
)

TEST_INIT_ARGVALS = itertools.product(LABELMAPS, INPUT_SHAPES)


class TestTweetyNet:

    @pytest.mark.parametrize(
        'labelmap, input_shape',
        TEST_INIT_ARGVALS
    )
    def test_init(self, labelmap, input_shape):
        # network has required args that need to be determined dynamically
        num_input_channels = input_shape[-3]
        num_freqbins = input_shape[-2]
        network = vak.models.TweetyNet.definition.network(len(labelmap), num_input_channels, num_freqbins)
        model = vak.models.TweetyNet.from_instances(network=network, labelmap=labelmap)
        assert isinstance(model, vak.models.FrameClassificationModel)
        for attr in ('network', 'loss', 'optimizer'):
            assert hasattr(model, attr)
            assert isinstance(getattr(model, attr),
                              getattr(vak.models.tweetynet.TweetyNet.definition, attr))
        assert hasattr(model, 'metrics')
        assert isinstance(model.metrics, dict)
        for metric_name, metric_callable in model.metrics.items():
            assert isinstance(metric_callable,
                              vak.models.tweetynet.TweetyNet.definition.metrics[metric_name])



import pytest

import vak.models

from .conftest import MockModel, MockModelFamily, MockEncoderDecoderModel
from .test_definition import TweetyNetDefinition as TweetyNet

from .test_definition import (
    MissingClassVarModelDefinition,
    ExtraClassVarModelDefinition,
    InvalidNetworkTypeModelDefinition,
    InvalidNetworkDictKeyModelDefinition,
    InvalidNetworkDictValueModelDefinition,
    InvalidLossTypeModelDefinition,
    InvalidOptimTypeModelDefinition,
    InvalidMetricsTypeModelDefinition,
    InvalidMetricsDictKeyModelDefinition,
    InvalidMetricsDictValueModelDefinition,
)


TweetyNet.__name__ = 'TweetyNet'


@pytest.mark.parametrize(
    'definition, family, expected_name',
    [
        (MockModel,
         MockModelFamily,
         'MockModel'),
        (MockEncoderDecoderModel,
         MockModelFamily,
         'MockEncoderDecoderModel'),
    ]
)
def test_model(definition, family, expected_name):
    """Test that :func:`vak.models.decorator.model` decorator
    returns a new instance of ModelFactory,
    and has the expected name"""
    model_class = vak.models.decorator.model(family)(definition)

    assert isinstance(model_class, vak.models.factory.ModelFactory)
    assert model_class.__name__ == expected_name

    # need to delete model from registry so other tests don't fail
    del vak.models.registry.MODEL_REGISTRY[model_class.__name__]


@pytest.mark.parametrize(
    'definition',
    [
        MissingClassVarModelDefinition,
        ExtraClassVarModelDefinition,
        InvalidNetworkTypeModelDefinition,
        InvalidNetworkDictKeyModelDefinition,
        InvalidNetworkDictValueModelDefinition,
        InvalidLossTypeModelDefinition,
        InvalidOptimTypeModelDefinition,
        InvalidMetricsTypeModelDefinition,
        InvalidMetricsDictKeyModelDefinition,
        InvalidMetricsDictValueModelDefinition,
    ]
)
def test_model_raises(definition):
    with pytest.raises(vak.models.decorator.ModelDefinitionValidationError):
        vak.models.decorator.model(MockModelFamily)(definition)



import json
import shutil
from unittest import mock

import numpy as np
import pandas as pd
import pytest

import vak.common.converters
import vak.common.labels
import vak.common.paths
import vak.prep.frame_classification

@pytest.mark.parametrize(
    'model_name, audio_format, annot_format, input_type',
    [
        ('TweetyNet', 'cbin', 'notmat', 'spect')
    ]
)
def test_make_index_vectors_for_each_subsets(
    model_name, audio_format, annot_format, input_type, specific_config_toml_path, trainer_table, tmp_path,
):
    root_results_dir = tmp_path.joinpath("tmp_root_results_dir")
    root_results_dir.mkdir()
    keys_to_change = [
        {
            "table": "learncurve",
            "key": "root_results_dir",
            "value": str(root_results_dir),
        },
    ]
    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    dataset_path = cfg.learncurve.dataset.path
    metadata = vak.datapipes.frame_classification.Metadata.from_dataset_path(dataset_path)
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)

    subsets_df = dataset_df[
        ~dataset_df['subset'].isnull()
    ]

    tmp_dataset_path = tmp_path / f"test_make_learncurve_splits_from_dataset_df"
    shutil.copytree(dataset_path, tmp_dataset_path)
    # delete all the subset indices vectors, since we're about to test that we make them
    for train_dur in cfg.prep.train_set_durs:
        for replicate_num in range(1, cfg.prep.num_replicates + 1):
            train_dur_replicate_subset_name = vak.common.learncurve.get_train_dur_replicate_subset_name(
                    train_dur, replicate_num
                )
            sample_id_vec_path = (tmp_dataset_path / "train" /
                                  vak.datapipes.frame_classification.helper.sample_ids_array_filename_for_subset(
                                      train_dur_replicate_subset_name)
                                  )
            sample_id_vec_path.unlink()
            inds_in_sample_vec_path = (tmp_dataset_path / "train" /
                                       vak.datapipes.frame_classification.helper.inds_in_sample_array_filename_for_subset(
                                       train_dur_replicate_subset_name)
                                  )
            inds_in_sample_vec_path.unlink()

    vak.prep.frame_classification.learncurve.make_index_vectors_for_each_subset(
        subsets_df,
        tmp_dataset_path,
        input_type,
    )

    assert sorted(subsets_df['train_dur'].unique()) == cfg.prep.train_set_durs
    assert sorted(
        subsets_df['replicate_num'].unique()
    ) == list(range(1, cfg.prep.num_replicates + 1))

    # assert that each expected split name is in data frame
    for train_dur in cfg.prep.train_set_durs:
        train_dur_df = subsets_df[np.isclose(subsets_df['train_dur'], train_dur)].copy()
        # assert correct number of replicates for this train duration
        assert sorted(
            train_dur_df['replicate_num']
        ) == list(range(1, cfg.prep.num_replicates + 1))

        for replicate_num in range(1, cfg.prep.num_replicates + 1):
            subset_name = vak.common.learncurve.get_train_dur_replicate_subset_name(
                    train_dur, replicate_num
                )

            # test that indexing vectors got made
            sample_id_vec_path = (tmp_dataset_path / "train" /
                                  vak.datapipes.frame_classification.helper.sample_ids_array_filename_for_subset(
                                      subset_name)
                                  )
            assert sample_id_vec_path.exists()

            inds_in_sample_vec_path = (tmp_dataset_path / "train" /
                                       vak.datapipes.frame_classification.helper.inds_in_sample_array_filename_for_subset(
                                           subset_name)
                                       )
            assert inds_in_sample_vec_path.exists()

            this_subset_df = subsets_df[subsets_df['subset'] == subset_name]
            frames_paths = this_subset_df[
                vak.datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME
            ].values
            sample_id_vec, inds_in_sample_vec = [], []
            for sample_id, frames_path in enumerate(frames_paths):
                # make indexing vectors that we use to test
                frames = vak.datapipes.frame_classification.helper.load_frames(tmp_dataset_path / frames_path,
                                                                              input_type)
                n_frames = frames.shape[-1]
                sample_id_vec.append(np.ones((n_frames,)).astype(np.int32) * sample_id)
                inds_in_sample_vec.append(np.arange(n_frames))
            expected_sample_id_vec = np.concatenate(sample_id_vec)
            expected_inds_in_sample_vec = np.concatenate(inds_in_sample_vec)
            sample_id_vec = np.load(sample_id_vec_path)
            assert np.array_equal(sample_id_vec, expected_sample_id_vec)
            inds_in_sample_vec = np.load(inds_in_sample_vec_path)
            assert np.array_equal(inds_in_sample_vec, expected_inds_in_sample_vec)


@pytest.mark.parametrize(
    'model_name, audio_format, annot_format, input_type',
    [
        ('TweetyNet', 'cbin', 'notmat', 'spect')
    ]
)
def test_make_subsets_from_dataset_df(
    model_name, audio_format, annot_format, input_type, specific_config_toml_path, trainer_table, tmp_path,
):
    root_results_dir = tmp_path.joinpath("tmp_root_results_dir")
    root_results_dir.mkdir()
    keys_to_change = [
        {
            "table": "learncurve",
            "key": "root_results_dir",
            "value": str(root_results_dir),
        },
    ]
    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    dataset_path = cfg.learncurve.dataset.path
    metadata = vak.datapipes.frame_classification.Metadata.from_dataset_path(dataset_path)
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)

    labelmap_path = dataset_path / "labelmap.json"
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)

    tmp_dataset_path = tmp_path / f"test_make_learncurve_splits_from_dataset_df"
    shutil.copytree(dataset_path, tmp_dataset_path)
    # delete all the subset indices vectors, since we're about to test that we make them
    for train_dur in cfg.prep.train_set_durs:
        for replicate_num in range(1, cfg.prep.num_replicates + 1):
            train_dur_replicate_subset_name = vak.common.learncurve.get_train_dur_replicate_subset_name(
                    train_dur, replicate_num
                )
            sample_id_vec_path = (tmp_dataset_path / "train" /
                                  vak.datapipes.frame_classification.helper.sample_ids_array_filename_for_subset(
                                      train_dur_replicate_subset_name)
                                  )
            sample_id_vec_path.unlink()
            inds_in_sample_vec_path = (tmp_dataset_path / "train" /
                                       vak.datapipes.frame_classification.helper.inds_in_sample_array_filename_for_subset(
                                       train_dur_replicate_subset_name)
                                  )
            inds_in_sample_vec_path.unlink()

    # now reset the dataset df to what it would have been before we passed it into `make_splits`
    dataset_df = dataset_df[
        # drop any rows where there *is* a train dur -- because these are the subsets
        dataset_df['train_dur'].isnull()
        # drop the columns added by ``make_splits``, then reset the index
    ].drop(columns=['subset', 'train_dur', 'replicate_num']).reset_index(drop=True)

    with mock.patch('vak.prep.frame_classification.learncurve.make_index_vectors_for_each_subset') as mock_idx_vectors:
        out = vak.prep.frame_classification.learncurve.make_subsets_from_dataset_df(
            dataset_df,
            input_type,
            cfg.prep.train_set_durs,
            cfg.prep.num_replicates,
            tmp_dataset_path,
            labelmap,
        )
        assert mock_idx_vectors.called

    assert isinstance(out, pd.DataFrame)

    for added_column in ('subset', 'train_dur', 'replicate_num'):
        assert added_column in out.columns

    subsets_df = out[
        ~out['subset'].isnull()
    ]
    assert sorted(subsets_df['train_dur'].unique()) == cfg.prep.train_set_durs
    assert sorted(
        subsets_df['replicate_num'].unique()
    ) == list(range(1, cfg.prep.num_replicates + 1))

    # assert that each expected split name is in data frame
    for train_dur in cfg.prep.train_set_durs:
        train_dur_df = subsets_df[np.isclose(subsets_df['train_dur'], train_dur)].copy()
        # assert correct number of replicates for this train duration
        assert sorted(
            train_dur_df['replicate_num']
        ) == list(range(1, cfg.prep.num_replicates + 1))

        for replicate_num in range(1, cfg.prep.num_replicates + 1):
            subset_name = vak.common.learncurve.get_train_dur_replicate_subset_name(
                    train_dur, replicate_num
                )

            # assert this train_dur + replicate split exists in dataframe
            assert np.isin(subset_name, subsets_df['subset'].values)
            this_subset_df = subsets_df[subsets_df['subset'] == subset_name]

            # assert that it has the correct duration
            assert this_subset_df['duration'].sum() >= train_dur



"""The :mod:`vak.common` module contains helper or utility functions
that are used in multiple other modules, e.g., validators
or logging functions.

If a helper/utility function is only used in one module,
it should live either in that module or another at the same level.
See for example :mod:`vak.prep.prep_helper` or
:mod:`vak.datsets.train_datapipe._helper`.
"""

from . import (
    accelerator,
    annotation,
    constants,
    converters,
    files,
    labels,
    learncurve,
    logging,
    paths,
    tensorboard,
    timebins,
    timenow,
    typing,
    validators,
)

__all__ = [
    "annotation",
    "constants",
    "converters",
    "accelerator",
    "files",
    "labels",
    "learncurve",
    "logging",
    "paths",
    "tensorboard",
    "timebins",
    "timenow",
    "typing",
    "validators",
]



"""fixtures relating to audio files"""
import pytest

from .annot import LABELSET_NOTMAT, ANNOT_LIST_NOTMAT
from .test_data import SOURCE_TEST_DATA_ROOT


@pytest.fixture
def default_spect_params():
    return dict(
        fft_size=512,
        step_size=64,
        freq_cutoffs=(500, 10000),
        thresh=6.25,
        transform_type="log_spect",
        freqbins_key="f",
        timebins_key="t",
        spect_key="s",
        audio_path_key="audio_path",
    )


AUDIO_DIR_CBIN = SOURCE_TEST_DATA_ROOT.joinpath("audio_cbin_annot_notmat", "gy6or6", "032312")


@pytest.fixture
def audio_dir_cbin():
    return AUDIO_DIR_CBIN


AUDIO_LIST_CBIN = sorted(AUDIO_DIR_CBIN.glob("*.cbin"))


@pytest.fixture
def audio_list_cbin():
    return AUDIO_LIST_CBIN

LABELSET_NOTMAT_AS_SET = set(LABELSET_NOTMAT)


AUDIO_LIST_CBIN_ALL_LABELS_IN_LABELSET = []
AUDIO_LIST_CBIN_LABELS_NOT_IN_LABELSET = []
for audio_path in AUDIO_LIST_CBIN:
    audio_fname = audio_path.name
    annot = [
        annot for annot in ANNOT_LIST_NOTMAT if annot.notated_path.name == audio_fname
    ]
    assert len(annot) == 1
    annot = annot[0]
    if set(annot.seq.labels).issubset(LABELSET_NOTMAT_AS_SET):
        AUDIO_LIST_CBIN_ALL_LABELS_IN_LABELSET.append(audio_path)
    else:
        AUDIO_LIST_CBIN_LABELS_NOT_IN_LABELSET.append(audio_path)

@pytest.fixture
def audio_list_cbin_all_labels_in_labelset():
    """list of .cbin audio files where all labels in associated annotation **are** in labelset"""
    return AUDIO_LIST_CBIN_ALL_LABELS_IN_LABELSET


@pytest.fixture
def audio_list_cbin_labels_not_in_labelset():
    """list of .cbin audio files where some labels in associated annotation are **not** in labelset"""
    return AUDIO_LIST_CBIN_LABELS_NOT_IN_LABELSET


@pytest.fixture
def audio_list_factory(audio_list_cbin):
    """factory fixture, returns a function that
    returns a fixture containing a list of Annotation objects,
    given a specified annotation format

    so that unit tests can be parameterized with annotation format names
    """
    FORMAT_AUDIO_LIST_FIXTURE_MAP = {
        "audio_cbin_annot_notmat": audio_list_cbin,
    }

    def _audio_list_factory(audio_format, annot_format):
        key = f'audio_{audio_format}_annot_{annot_format}'
        return FORMAT_AUDIO_LIST_FIXTURE_MAP[key]

    return _audio_list_factory


@pytest.fixture
def specific_audio_list(
    audio_list_cbin,
    audio_list_cbin_all_labels_in_labelset,
    audio_list_cbin_labels_not_in_labelset,
):
    def _specific_audio_list(spect_format, qualifier=None):
        MAP = {
            "cbin": {
                None: audio_list_cbin,
                "all_labels_in_labelset": audio_list_cbin_all_labels_in_labelset,
                "labels_not_in_labelset": audio_list_cbin_labels_not_in_labelset,
            },
        }
        return MAP[spect_format][qualifier]

    return _specific_audio_list



import inspect
import itertools

import pytest
import torch

import vak

from .conftest import (
    MockAcc,
    MockDecoder,
    MockEncoder,
    MockEncoderDecoderModel,
    MockModel,
    MockModelFamily,
    MockNetwork,
    other_loss_func,
    other_metrics_dict,
    OtherNetwork,
    OtherOptimizer,
)

from .test_definition import (
    InvalidMetricsDictKeyModelDefinition,
    TweetyNetDefinition,
)
from .test_tweetynet import LABELMAPS, INPUT_SHAPES

MODEL_DEFINITION_CLASS_VARS = (
    'network',
    'loss',
    'optimizer',
    'metrics',
    'default_config'
)

mock_net_instance = MockNetwork()

TEST_VALIDATE_RAISES_ARGVALS = [
    (MockModel, dict(network=OtherNetwork()), TypeError),
    (MockModel, dict(loss=other_loss_func), TypeError),
    (MockModel, dict(optimizer=OtherOptimizer), TypeError),
    (MockModel, dict(metrics=other_metrics_dict), ValueError),
    (MockEncoderDecoderModel,
    # first value is wrong
     dict(network={'MockEncoder': OtherNetwork(), 'MockDecoder': MockDecoder()}),
     TypeError),
    (MockEncoderDecoderModel,
     # missng key, MockEncoder
     dict(network={'MockDecoder': MockDecoder()}),
     ValueError),
    (MockEncoderDecoderModel,
     # extra key, MockRecoder
     dict(network={'MockEncoder': MockEncoder(), 'MockDecoder': MockDecoder(), 'MockRecoder': MockNetwork()}),
     ValueError),
]

# pytest.mark.parametrize vals for test_init_with_definition
MODEL_DEFS = (
    TweetyNetDefinition,
)

TEST_WITH_FRAME_CLASSIFICATION_ARGVALS = itertools.product(LABELMAPS, INPUT_SHAPES, MODEL_DEFS)

MOCK_INPUT_SHAPE = torch.Size([1, 128, 44])


class ConvEncoderUMAPDefinition:
    network = {"encoder": vak.nets.ConvEncoder}
    loss = vak.nn.UmapLoss
    optimizer = torch.optim.AdamW
    metrics = {
        "acc": vak.metrics.Accuracy,
        "levenshtein": vak.metrics.Levenshtein,
        "character_error_rate": vak.metrics.CharacterErrorRate,
        "loss": torch.nn.CrossEntropyLoss,
    }
    default_config = {
        "optimizer": {"lr": 1e-3},
    }


class TestModelFactory:
    def test_init_no_definition_raises(self):
        """Test that initializing a Model instance without a definition or family raises a ValueError."""
        with pytest.raises(TypeError):
            vak.models.factory.ModelFactory()

    def test_init_invalid_definition_raises(self):
        """Test that initializing a Model instance with an invalid definition raises a ValueError."""
        with pytest.raises(vak.models.decorator.ModelDefinitionValidationError):
            vak.models.factory.ModelFactory(
                definition=InvalidMetricsDictKeyModelDefinition,
                family=MockModelFamily,
            )

    @pytest.mark.parametrize(
        'definition, kwargs',
        [
            (MockModel, None),
            (MockModel, {'network': mock_net_instance}),
            (MockModel, {'loss': torch.nn.CrossEntropyLoss()},),
            (MockModel,
            {
                'network': mock_net_instance,
                'optimizer': torch.optim.SGD(lr=0.003, params=mock_net_instance.parameters())
            }
            ),
            (MockModel,
            {'metrics':
                {
                    'acc': MockAcc(),
                }
            }),
            (MockEncoderDecoderModel, None),
        ]
    )
    def test_validate_instances_or_get_default(self, definition, kwargs):
        model = vak.models.factory.ModelFactory(
            definition,
            MockModelFamily,
        )
        # actually instantiate model
        if kwargs:
            (network,
            loss,
            optimizer,
            metrics
            ) = model.validate_instances_or_get_default(**kwargs)
        else:
            (network,
            loss,
            optimizer,
            metrics
            ) = model.validate_instances_or_get_default()

        model_attrs = {
            'network': network,
            'loss': loss,
            'optimizer': optimizer,
            'metrics': metrics,
        }
        for attr in ('network', 'loss', 'optimizer', 'metrics'):
            model_attr = model_attrs[attr]
            definition_attr = getattr(definition, attr)
            if inspect.isclass(definition_attr):
                assert isinstance(model_attr, definition_attr)
            elif isinstance(definition_attr, dict):
                assert isinstance(model_attr, dict)
                for definition_key, definition_val in definition_attr.items():
                    assert definition_key in model_attr
                    model_val = model_attr[definition_key]
                    if inspect.isclass(definition_val):
                        assert isinstance(model_val, definition_val)
                    else:
                        assert callable(definition_val)
                        assert model_val is definition_val
            else:
                # must be a function
                assert callable(model_attr)
                assert model_attr is definition_attr

    @pytest.mark.parametrize(
        'definition, kwargs, expected_exception',
        TEST_VALIDATE_RAISES_ARGVALS
    )
    def test_validate_instances_or_get_default_raises(self, definition, kwargs, expected_exception):
        """Test that :meth:`validate_instances_or_get_default` raises errors as expected given input arguments.

        Note that this should happen from ``validate_instances_or_get_default`` calling ``validate_init``,
        so here we test that this is happening inside ``validate_instances_or_get_default``.
        Next method tests ``validate_init`` directly.
        """
        model = vak.models.factory.ModelFactory(
            definition,
            MockModelFamily,
        )
        with pytest.raises(expected_exception):
            model.validate_instances_or_get_default(**kwargs)

    @pytest.mark.parametrize(
        'definition, kwargs, expected_exception',
        TEST_VALIDATE_RAISES_ARGVALS
    )
    def test_validate_init_raises(self, definition, kwargs, expected_exception):
        """Test that ``validate_init`` raises errors as expected"""
        model = vak.models.factory.ModelFactory(
            definition=definition,
            family=MockModelFamily
        )
        with pytest.raises(expected_exception):
            model.validate_init(**kwargs)

    MODEL_DEFINITION_MAP = {
        'TweetyNet': TweetyNetDefinition,
    }

    @pytest.mark.parametrize(
        'definition, config',
        [
            (MockModel, {'network': {'n_classes': 10}}),
            (MockModel, {'loss': {'reduction': 'sum'}}),
            (MockModel, {
                 'network': {'n_classes': 10},
                 'optimizer': {'lr': 0.003}}
             ),
            (MockModel, {'metrics': {'acc': {'average': 'micro'}}}),
            (MockEncoderDecoderModel, {
                 'network': {
                     'MockEncoder': {'input_size': 5},
                     'MockDecoder': {'output_size': 5},
                 },
                 'optimizer': {'lr': 0.003}}
             )
        ]
    )
    def test_from_config(self,
                         definition,
                         config,
                         ):
        model = vak.models.factory.ModelFactory(
            definition=definition,
            family=MockModelFamily
        )
        new_model_instance = model.from_config(config)

        assert isinstance(new_model_instance, MockModelFamily)

        if 'network' in config:
            if inspect.isclass(definition.network):
                for network_kwarg, network_kwargval in config['network'].items():
                    assert hasattr(new_model_instance.network, network_kwarg)
                    assert getattr(new_model_instance.network, network_kwarg) == network_kwargval
            elif isinstance(definition.network, dict):
                for net_name, net_kwargs in config['network'].items():
                    for network_kwarg, network_kwargval in net_kwargs.items():
                        assert hasattr(new_model_instance.network[net_name], network_kwarg)
                        assert getattr(new_model_instance.network[net_name], network_kwarg) == network_kwargval

        if 'loss' in config:
            for loss_kwarg, loss_kwargval in config['loss'].items():
                assert hasattr(new_model_instance.loss, loss_kwarg)
                assert getattr(new_model_instance.loss, loss_kwarg) == loss_kwargval

        if 'optimizer' in config:
            for optimizer_kwarg, optimizer_kwargval in config['optimizer'].items():
                assert optimizer_kwarg in new_model_instance.optimizer.param_groups[0]
                assert new_model_instance.optimizer.param_groups[0][optimizer_kwarg] == optimizer_kwargval

        if 'metrics' in config:
            for metric_name, metric_kwargs in config['metrics'].items():
                assert metric_name in new_model_instance.metrics
                for metric_kwarg, metric_kwargval in metric_kwargs.items():
                    assert hasattr(new_model_instance.metrics[metric_name], metric_kwarg)
                    assert getattr(new_model_instance.metrics[metric_name], metric_kwarg) == metric_kwargval

    @pytest.mark.parametrize(
            'labelmap, input_shape, definition',
            TEST_WITH_FRAME_CLASSIFICATION_ARGVALS
    )
    def test_from_config_frame_classification(self, labelmap, input_shape, definition):
        model_factory = vak.models.factory.ModelFactory(
            definition,
            vak.models.FrameClassificationModel,
        )
        num_input_channels, num_freqbins = input_shape[0], input_shape[1]
        # network has required args that need to be determined dynamically
        network = definition.network(len(labelmap), num_input_channels, num_freqbins)
        model = model_factory.from_instances(network=network, labelmap=labelmap)

        # now test that attributes are what we expect
        assert isinstance(model, vak.models.FrameClassificationModel)
        for attr in ('network', 'loss', 'optimizer', 'metrics'):
            assert hasattr(model, attr)
            model_attr = getattr(model, attr)
            definition_attr = getattr(definition, attr)
            if inspect.isclass(definition_attr):
                assert isinstance(model_attr, definition_attr)
            elif isinstance(definition_attr, dict):
                assert isinstance(model_attr, dict)
                for definition_key, definition_val in definition_attr.items():
                    assert definition_key in model_attr
                    model_val = model_attr[definition_key]
                    if inspect.isclass(definition_val):
                        assert isinstance(model_val, definition_val)
                    else:
                        assert callable(definition_val)
                        assert model_val is definition_val
            else:
                # must be a function
                assert callable(model_attr)
                assert model_attr is definition_attr

    @pytest.mark.parametrize(
        'definition',
        [
            TweetyNetDefinition,
        ]
    )
    def test_from_config_with_frame_classification(self, definition, specific_config_toml_path):
        model_name = definition.__name__.replace('Definition', '')
        toml_path = specific_config_toml_path('train', model_name, audio_format='cbin', annot_format='notmat')
        cfg = vak.config.Config.from_toml_path(toml_path)

        # stuff we need just to be able to instantiate network
        labelmap = vak.common.labels.to_map(cfg.prep.labelset, map_background=True)

        model_factory = vak.models.factory.ModelFactory(
            definition,
            vak.models.FrameClassificationModel,
        )

        config = cfg.train.model.asdict()
        num_input_channels, num_freqbins = MOCK_INPUT_SHAPE[0], MOCK_INPUT_SHAPE[1]

        config["network"].update(
            num_classes=len(labelmap),
            num_input_channels=num_input_channels,
            num_freqbins=num_freqbins
        )

        model = model_factory.from_config(config=config, labelmap=labelmap)
        assert isinstance(model, vak.models.FrameClassificationModel)

        # below, we can only test the config kwargs that actually end up as attributes
        # so we use `if hasattr` before checking
        if 'network' in config:
            if inspect.isclass(definition.network):
                for network_kwarg, network_kwargval in config['network'].items():
                    if hasattr(model.network, network_kwarg):
                        assert getattr(model.network, network_kwarg) == network_kwargval
            elif isinstance(definition.network, dict):
                for net_name, net_kwargs in config['network'].items():
                    for network_kwarg, network_kwargval in net_kwargs.items():
                        if hasattr(model.network[net_name], network_kwarg):
                            assert getattr(model.network[net_name], network_kwarg) == network_kwargval

        if 'loss' in config:
            for loss_kwarg, loss_kwargval in config['loss'].items():
                if hasattr(model.loss, loss_kwarg):
                    assert getattr(model.loss, loss_kwarg) == loss_kwargval

        if 'optimizer' in config:
            for optimizer_kwarg, optimizer_kwargval in config['optimizer'].items():
                if optimizer_kwarg in model.optimizer.param_groups[0]:
                    assert model.optimizer.param_groups[0][optimizer_kwarg] == optimizer_kwargval

        if 'metrics' in config:
            for metric_name, metric_kwargs in config['metrics'].items():
                assert metric_name in model.metrics
                for metric_kwarg, metric_kwargval in metric_kwargs.items():
                    if hasattr(model.metrics[metric_name], metric_kwarg):
                        assert getattr(model.metrics[metric_name], metric_kwarg) == metric_kwargval

    @pytest.mark.parametrize(
        'input_shape, definition',
        [
            ((1, 128, 128), ConvEncoderUMAPDefinition),
        ]
    )
    def test_from_instances_parametric_umap(
            self,
            input_shape,
            definition,
    ):
        network = {'encoder': vak.nets.ConvEncoder(input_shape)}

        model_factory = vak.models.ModelFactory(
            definition,
            vak.models.ParametricUMAPModel,
        )
        model = model_factory.from_instances(network=network)

        # now test that attributes are what we expect
        assert isinstance(model, vak.models.ParametricUMAPModel)
        for attr in ('network', 'loss', 'optimizer', 'metrics'):
            assert hasattr(model, attr)
            model_attr = getattr(model, attr)
            definition_attr = getattr(definition, attr)
            if inspect.isclass(definition_attr):
                assert isinstance(model_attr, definition_attr)
            elif isinstance(definition_attr, dict):
                assert isinstance(model_attr, (dict, torch.nn.ModuleDict))
                for definition_key, definition_val in definition_attr.items():
                    assert definition_key in model_attr
                    model_val = model_attr[definition_key]
                    if inspect.isclass(definition_val):
                        assert isinstance(model_val, definition_val)
                    else:
                        assert callable(definition_val)
                        assert model_val is definition_val
            else:
                # must be a function
                assert callable(model_attr)
                assert model_attr is definition_attr

    @pytest.mark.parametrize(
        'input_shape, definition',
        [
            ((1, 128, 128), ConvEncoderUMAPDefinition),
        ]
    )
    def test_from_config_with_parametric_umap(
            self,
            input_shape,
            definition,
            specific_config_toml_path,
    ):
        model_name = definition.__name__.replace('Definition', '')
        toml_path = specific_config_toml_path('train', model_name, audio_format='cbin', annot_format='notmat')
        cfg = vak.config.Config.from_toml_path(toml_path)

        model_factory = vak.models.ModelFactory(
            definition,
            vak.models.ParametricUMAPModel,
        )

        config = cfg.train.model.asdict()
        config["network"]["encoder"]["input_shape"] = input_shape

        model = model_factory.from_config(config=config)
        assert isinstance(model, vak.models.ParametricUMAPModel)

        if 'network' in config:
            if inspect.isclass(definition.network):
                for network_kwarg, network_kwargval in config['network'].items():
                    assert hasattr(model.network, network_kwarg)
                    assert getattr(model.network, network_kwarg) == network_kwargval
            elif isinstance(definition.network, dict):
                for net_name, net_kwargs in config['network'].items():
                    for network_kwarg, network_kwargval in net_kwargs.items():
                        network = model.network[net_name]
                        if hasattr(network, network_kwarg):
                            assert getattr(network, network_kwarg) == network_kwargval

        if 'loss' in config:
            for loss_kwarg, loss_kwargval in config['loss'].items():
                assert hasattr(model.loss, loss_kwarg)
                assert getattr(model.loss, loss_kwarg) == loss_kwargval

        if 'optimizer' in config:
            for optimizer_kwarg, optimizer_kwargval in config['optimizer'].items():
                assert optimizer_kwarg in model.optimizer.param_groups[0]
                assert model.optimizer.param_groups[0][optimizer_kwarg] == optimizer_kwargval

        if 'metrics' in config:
            for metric_name, metric_kwargs in config['metrics'].items():
                assert metric_name in model.metrics
                for metric_kwarg, metric_kwargval in metric_kwargs.items():
                    assert hasattr(model.metrics[metric_name], metric_kwarg)
                    assert getattr(model.metrics[metric_name], metric_kwarg) == metric_kwargval



import pytest

import vak.config.model


class TestModelConfig:

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'NonExistentModel': {
                        'network': {},
                        'optimizer': {},
                        'loss': {},
                        'metrics': {},
                    }
                },
                {
                    'TweetyNet': {
                        'network': {},
                        'optimizer': {'lr': 1e-3},
                        'loss': {},
                        'metrics': {},
                    }
                },
            ]
    )
    def test_init(self, config_dict):
        name=list(config_dict.keys())[0]
        config_dict_from_name = config_dict[name]

        model_config = vak.config.model.ModelConfig(
            name=name,
            **config_dict_from_name
            )

        assert isinstance(model_config, vak.config.model.ModelConfig)
        assert model_config.name == name
        for key, val in config_dict_from_name.items():
            assert hasattr(model_config, key)
            assert getattr(model_config, key) == val

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'TweetyNet': {
                        'optimizer': {'lr': 1e-3},
                    }
                },
                {
                    'TweetyNet': {
                        'network': {},
                        'optimizer': {'lr': 1e-3},
                        'loss': {},
                        'metrics': {},
                    }
                },
                {
                    'ED_TCN': {
                        'optimizer': {'lr': 1e-3},
                    }
                },
                {
                    "ConvEncoderUMAP": {
                        "optimizer": {'lr': 1e-3},
                    }
                },
                {
                    "ConvEncoderUMAP": {
                        "network": {
                            "encoder": {
                                "conv1_filters": 8,
                            }
                        },
                        "optimizer": {'lr': 1e-3},
                    }
                }
            ]
    )
    def test_from_config_dict(self, config_dict):
        model_config = vak.config.model.ModelConfig.from_config_dict(config_dict)

        name=list(config_dict.keys())[0]
        config_dict_from_name = config_dict[name]
        assert model_config.name == name
        for attr in ('network', 'optimizer', 'loss', 'metrics'):
            assert hasattr(model_config, attr)
            if attr in config_dict_from_name:
                assert getattr(model_config, attr) == config_dict_from_name[attr]
            else:
                assert getattr(model_config, attr) == {}

    def test_from_config_dict_real_config(self, a_generated_config_dict):
        config_dict = None
        for table_name in ('train', 'eval', 'predict', 'learncurve'):
            if table_name in a_generated_config_dict:
                config_dict = a_generated_config_dict[table_name]['model']
        if config_dict is None:
            raise ValueError(
                f"Didn't find top-level table for config: {a_generated_config_dict}"
            )

        model_config = vak.config.model.ModelConfig.from_config_dict(config_dict)

        name=list(config_dict.keys())[0]
        config_dict_from_name = config_dict[name]
        assert model_config.name == name
        for attr in ('network', 'optimizer', 'loss', 'metrics'):
            assert hasattr(model_config, attr)
            if attr in config_dict_from_name:
                assert getattr(model_config, attr) == config_dict_from_name[attr]
            else:
                assert getattr(model_config, attr) == {}

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'TweetyNet': {
                        'optimizer': {'lr': 1e-3},
                    }
                },
                {
                    'TweetyNet': {
                        'network': {},
                        'optimizer': {'lr': 1e-3},
                        'loss': {},
                        'metrics': {},
                    }
                },
                {
                    'ED_TCN': {
                        'optimizer': {'lr': 1e-3},
                    }
                },
                {
                    "ConvEncoderUMAP": {
                        "optimizer": {'lr': 1e-3},
                    }
                },
                {
                    "ConvEncoderUMAP": {
                        "network": {
                            "encoder": {
                                "conv1_filters": 8,
                            }
                        },
                        "optimizer": {'lr': 1e-3},
                    }
                }
            ]
    )
    def test_asdict(self, config_dict):
        model_config = vak.config.model.ModelConfig.from_config_dict(config_dict)

        model_config_as_dict = model_config.asdict()

        assert isinstance(model_config_as_dict, dict)

        model_name = list(config_dict.keys())[0]
        for key in ('name', 'network', 'optimizer', 'loss', 'metrics'):
            if key == 'name':
                assert model_config_as_dict[key] == model_name
            else:
                if key in config_dict[model_name]:
                    assert model_config_as_dict[key] == config_dict[model_name][key]
                else:
                    assert model_config_as_dict[key] == {}

    @pytest.mark.parametrize(
            'config_dict, expected_exception',
            [
                # Raises ValueError because model name not in registry
                (
                    {
                        'NonExistentModel': {
                            'network': {},
                            'optimizer': {},
                            'loss': {},
                            'metrics': {},
                        }
                    },
                    ValueError,
                )
            ]
    )
    def test_from_config_dict_raises(self, config_dict, expected_exception):
        with pytest.raises(expected_exception):
            vak.config.model.ModelConfig.from_config_dict(config_dict)



import pytest
import torch

import vak


class TestConvEncoderUMAP:
    @pytest.mark.parametrize(
        'input_shape',
        [
            (1, 32, 32),
            (1, 64, 64),
        ]
    )
    def test_from_instances(self, input_shape):
        network = {
            'encoder': vak.models.ConvEncoderUMAP.definition.network['encoder'](input_shape=input_shape)
        }
        model = vak.models.ConvEncoderUMAP.from_instances(network=network)
        assert isinstance(model, vak.models.ParametricUMAPModel)
        for attr in ('network', 'loss', 'optimizer'):
            assert hasattr(model, attr)
            attr_from_definition = getattr(vak.models.convencoder_umap.ConvEncoderUMAP.definition, attr)
            if isinstance(attr_from_definition, dict):
                attr_from_model = getattr(model, attr)
                assert isinstance(attr_from_model, (dict, torch.nn.ModuleDict))
                assert attr_from_model.keys() == attr_from_definition.keys()
                for net_name, net_instance in attr_from_model.items():
                    assert isinstance(net_instance, attr_from_definition[net_name])
            else:
                assert isinstance(getattr(model, attr),
                                  getattr(vak.models.convencoder_umap.ConvEncoderUMAP.definition, attr))
        assert hasattr(model, 'metrics')
        assert isinstance(model.metrics, dict)
        for metric_name, metric_callable in model.metrics.items():
            assert isinstance(metric_callable,
                              vak.models.convencoder_umap.ConvEncoderUMAP.definition.metrics[metric_name])



"""Helper functions for working with datasets represented as a pandas.DataFrame"""

from __future__ import annotations

import pathlib

import numpy as np
import pandas as pd


def get_dataset_csv_filename(data_dir_name: str, timenow: str) -> str:
    """Get name of csv file representing dataset.

    This function is called by
    :func:`vak.prep.frame_classification.dataset_df.get_dataset_csv_path`.

    Parameters
    ----------
    data_dir_name : str
        Name of directory specified as parameter ``data_dir``
        when calling :func:`vak.core.prep.prep`.
        This becomes the "prefix" of the csv filename.
    timenow : str
        Timestamp.
        This becomes the "suffix" of the csv filename.

    Returns
    -------
    dataset_csv_filename : str
        String, in the form f"{data_dir_name}_prep_{timenow}.csv"
    """
    return f"{data_dir_name}_prep_{timenow}.csv"


def get_dataset_csv_path(
    dataset_path: pathlib.Path, data_dir_name: str, timenow: str
) -> pathlib.Path:
    """Returns the path that should be used to save
    a pandas DataFrame representing a dataset
    to a csv file.

    Parameters
    ----------
    dataset_path : str, pathlib.Path
        Path to directory that represents dataset.
    data_dir_name : str
        Name of directory specified as parameter ``data_dir``
        when calling :func:`vak.core.prep.prep`.
        This becomes the "prefix" of the csv filename.
    timenow : str
        Timestamp.
        This becomes the "suffix" of the csv filename.

    Returns
    -------
    dataset_csv_path : pathlib.Path
        Path that is used when saving ``dataset_df`` as a csv file
        in the root of the dataset directory, ``dataset_path``.
    """
    dataset_csv_filename = get_dataset_csv_filename(data_dir_name, timenow)
    dataset_csv_path = dataset_path / dataset_csv_filename
    return dataset_csv_path


def add_split_col(df: pd.DataFrame, split: str) -> pd.DataFrame:
    """Add a 'split' column to a pandas DataFrame.

    Used by :func:`vak.prep`
    to assign an entire dataset to the same split,
    e.g. 'train' or 'predict'.
    All rows in the 'split' column will have the value specified.

    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe that represents a dataset.
    split : str
        A string that will be assigned to every row
        in the added "split" column.
        One of {'train', 'val', 'test', 'predict'}.
    """
    if split not in {"train", "val", "test", "predict"}:
        raise ValueError(
            f"value for split should be one of {{'train', 'val', 'test', 'predict'}}, but was '{split}'"
        )
    split_col = np.asarray([split for _ in range(len(df))], dtype="object")
    df["split"] = split_col
    return df



import pathlib
from unittest import mock

import pandas as pd
import pytest

import vak


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format, dataset_prep_func_to_mock",
    [
        ("eval", "cbin", None, "notmat", "vak.prep.prep_.prep_frame_classification_dataset"),
        ("learncurve", "cbin", None, "notmat", "vak.prep.prep_.prep_frame_classification_dataset"),
        ("predict", "cbin", None, "notmat", "vak.prep.prep_.prep_frame_classification_dataset"),
        ("train", "cbin", None, "notmat", "vak.prep.prep_.prep_frame_classification_dataset"),
        ("train", None, "mat", "yarden", "vak.prep.prep_.prep_frame_classification_dataset"),
    ],
)
def test_prep(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    dataset_prep_func_to_mock,
        specific_config_toml_path,
    default_model,
    tmp_path,
):
    # ---- set up
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {
            "table": "prep",
            "key": "output_dir",
            "value": str(output_dir),
        },
    ]
    toml_path = specific_config_toml_path(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    purpose = config_type.lower()

    # ---- test
    with mock.patch(dataset_prep_func_to_mock, autospec=True) as mocked_dataset_prep_func:
        mocked_dataset_prep_func.return_value = (pd.DataFrame.from_records([]), pathlib.Path('./fake/dataset/path'))
        _ = vak.prep.prep(
            data_dir=cfg.prep.data_dir,
            dataset_type=cfg.prep.dataset_type,
            input_type=cfg.prep.input_type,
            purpose=purpose,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            output_dir=cfg.prep.output_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
            train_set_durs=cfg.prep.train_set_durs,
            num_replicates=cfg.prep.num_replicates,
        )

        assert mocked_dataset_prep_func.called
